
__all__ = ["import_graphics_modules",
           "save_summary_plots",
           "create_overview_plot",
           "create_average_plot"]


def import_graphics_modules(verbose=False):
    """Imports pandas and matplotlib in a safe way, giving a good
       error message if something goes wrong.

       Parameters
       ----------
       verbose: bool
         Whether or not to print to the screen to signal progress...

       Returns
       -------
       (pd, plt)
          The pandas (pd) and matplotlib.pyplot (plt) modules
    """
    try:
        if verbose:
            print("Importing graphics modules...")
        import pandas as pd
        import matplotlib.pyplot as plt
    except ImportError:
        print("You must have pandas and matplotlib installed to run "
              "metawards-plot")
        print("Install using either `pip install pandas` if you are using")
        print("pip, or 'conda install pandas' if you are using conda.")
        raise ImportError("Cannot produce the plot as pandas and matplotlib "
                          "are not installed.")

    return (pd, plt)


def create_overview_plot(df, match_axes=True):
    """Create a summary plot of the result.csv data held in the
       passed pandas dataframe. This returns the figure for you
       to save if desired (or just call ``plt.show()`` to show
       it in Jupyter)

       If the dataframe contains multiple fingerprints, then this
       will return a dictionary of figures, one for each fingerprint,
       indexed by fingerprint

       Parameters
       ----------
       df : Pandas Dataframe
         The pandas dataframe containing the data from results.csv.bz2
       match_axes: bool
         If true (default) then this will ensure that all of the plots
         for different fingerprints are put on the same axis scale

       Returns
       -------
       fig
         The matplotlib figure containing the summary plot, or a
         dictionary of figures if there are multiple fingerprints
    """
    _, plt = import_graphics_modules()

    fingerprints = df["fingerprint"].unique()

    figs = {}

    min_date = None
    max_date = None
    max_y = {}
    min_y = {}

    columns = ["E", "I", "IW", "R"]

    if len(fingerprints) > 1 and match_axes:
        for fingerprint in fingerprints:
            df2 = df[df["fingerprint"] == fingerprint]

            for column in columns:
                min_d = df2["day"].min()
                max_d = df2["day"].max()
                min_val = df2[column].min()
                max_val = df2[column].max()

                if min_date is None:
                    min_date = min_d
                    max_date = max_d
                else:
                    if min_d < min_date:
                        min_date = min_d
                    if max_d > max_date:
                        max_date = max_d

                if column not in min_y:
                    min_y[column] = min_val
                    max_y[column] = max_val
                else:
                    if min_val < min_y[column]:
                        min_y[column] = min_val
                    if max_val > max_y[column]:
                        max_y[column] = max_val

    for fingerprint in fingerprints:
        df2 = df[df["fingerprint"] == fingerprint]

        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))

        i = 0
        j = 0

        for column in columns:
            ax = df2.pivot(index="date", columns="repeat",
                           values=column).plot.line(ax=axes[i][j])
            ax.tick_params('x', labelrotation=90)
            ax.get_legend().remove()
            ax.set_ylabel("Population")

            if len(fingerprints) > 1 and match_axes:
                ax.set_xlim(min_date, max_date)
                ax.set_ylim(min_y[column], 1.1*max_y[column])

            if len(fingerprints) > 1:
                ax.set_title(f"{fingerprint} : {column}")
            else:
                ax.set_title(column)

            j += 1
            if j == 2:
                j = 0
                i += 1

        fig.tight_layout(pad=1)
        figs[fingerprint] = fig

    if len(figs) == 0:
        return None
    elif len(figs) == 1:
        return figs[list(figs.keys())[0]]
    else:
        return figs


def create_average_plot(df):
    """Create an average plot of the result.csv data held in the
       passed pandas dataframe. This returns the figure for you
       to save if desired (or just call ``plt.show()`` to show
       it in Jupyter)

       Note that this won't do anything unless there are multiple
       repeats of the model run in the output. In that case, it
       will return None

       If the dataframe contains multiple fingerprints, then this
       will return a dictionary of figures, one for each fingerprint,
       indexed by fingerprint

       Parameters
       ----------
       df : Pandas Dataframe
         The pandas dataframe containing the data from results.csv.bz2

       Returns
       -------
       fig
         The matplotlib figure containing the summary plot, or None
         if there are no repeats over which to average. A dictionary
         of figures will be returned if the dataframe contains multiple
         fingerprint - the dictionaries will be indexed by fingerprint
    """
    fingerprints = df["fingerprint"].unique()

    figs = {}

    for fingerprint in fingerprints:
        df2 = df[df["fingerprint"] == fingerprint]

        nrepeats = len(df2["repeat"].unique())

        if nrepeats > 1:
            _, plt = import_graphics_modules()

            fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(10, 10))

            mean_average = df2.groupby("date").mean()
            stddev = df2.groupby("date").std()

            i = 0
            j = 0

            for column in ["E", "I", "IW", "R"]:
                ax = mean_average.plot.line(y=column, yerr=stddev[column],
                                            ax=axes[i][j])
                ax.tick_params('x', labelrotation=90)
                ax.get_legend().remove()
                ax.set_title(column)
                ax.set_ylabel("Population")

                j += 1
                if j == 2:
                    j = 0
                    i += 1

            fig.tight_layout(pad=1)
            figs[fingerprint] = fig

    if len(figs) == 0:
        return None
    elif len(figs) == 1:
        return figs[list(figs.keys())[0]]
    else:
        return figs


def save_summary_plots(results: str, output_dir: str = None,
                       format: str = "pdf", dpi: int = 300,
                       verbose=False):
    """Create summary plots of the data contained in the passed
       'results.csv.bz2' file that was produced by metawards
       and save them to disk.

       Parameters
       ----------
       results: str
         The full path to the file containing the results. This
         **must** have been created by ``metawards``
       output_dir: str
         Path to the directory in which you want to place the graphs.
         This defaults to the same directory that contains 'results'
       format: str
         The format to use to save the graphs. This defaults to 'pdf'
       dpi: int
         The dots-per-inch to use when saving bitmap graphics (e.g.
         png, jpg etc)
       verbose: bool
         Whether or not to print progress to the screen

       Returns
       -------
       filenames: List(str)
         Full file paths of all of the files written by this function
    """
    pd, _ = import_graphics_modules(verbose=verbose)
    import os

    if verbose:
        print(f"Reading data from {results}...")

    df = pd.read_csv(results)

    if output_dir is None:
        output_dir = os.path.dirname(results)

    if format is None:
        format = "pdf"

    filenames = []

    if verbose:
        print(f"Creating overview plot(s)...")

    figs = create_overview_plot(df)

    if not isinstance(figs, dict):
        figs = {"fingerprint": figs}

    for fingerprint, fig in figs.items():
        if len(figs) == 1:
            filename = os.path.join(output_dir, f"overview.{format}")
        else:
            filename = os.path.join(output_dir,
                                    f"overview_{fingerprint}.{format}")

        if verbose:
            print(f"Saving to {filename}...")

        fig.savefig(filename, dpi=dpi)
        filenames.append(filename)

    if verbose:
        print(f"Creating average plot(s)...")

    figs = create_average_plot(df)

    if figs is None:
        print("Nothing to plot")
    else:
        if not isinstance(figs, dict):
            figs = {"fingerprint": figs}

        for fingerprint, fig in figs.items():
            if len(figs) == 1:
                filename = os.path.join(output_dir, f"average.{format}")
            else:
                filename = os.path.join(output_dir,
                                        f"average_{fingerprint}.{format}")

            if verbose:
                print(f"Saving to {filename}...")

            fig.savefig(filename, dpi=dpi)
            filenames.append(filename)

    return filenames




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "branch": "master",
 "date": "2020-04-17T18:53:27+0100",
 "dirty": false,
 "error": null,
 "full-revisionid": "7f6feed543105843e5967c7ccbeee9badfb78f5a",
 "repository": "https://github.com/metawards/MetaWards",
 "version": "0.7.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)




__all__ = ["iterate_working_week"]

from ._iterate_core import iterate_core
from ._iterate_weekday import iterate_weekday
from ._iterate_weekend import iterate_weekend
from ._iterate_default import iterator_needs_setup

from .._population import Population


def iterate_working_week(population: Population,
                         setup=False, **kwargs):
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence for each iteration of the model run.
       This iterator understands the concept of a traditional working week,
       namely Monday-Friday is a work day, while Saturday and
       Sunday are weekends

       Parameters
       ----------
       population: Population
         The population experiencing the outbreak. This includes
         information about the day and date of the outbreak
       setup: bool
         Whether or not to return the functions used to setup the
         space and input for the advance_XXX functions returned by
         this iterator. This is called once at the start of a run
         to return the functions that must be called to setup the
         model

       Returns
       -------
       funcs: List[function]
         The list of functions that ```iterate``` will call in sequence
    """

    kwargs["population"] = population
    kwargs["setup"] = setup

    if setup:
        # Return the functions needed to initialise this iterator
        funcs = iterate_core(**kwargs)

        # Does 'iterate_weekday' need to be setup?
        if iterator_needs_setup(iterate_weekday):
            funcs += iterate_weekday(**kwargs)

        # Does 'iterate_weekend' need to be setup?
        if iterator_needs_setup(iterate_weekend):
            funcs += iterate_weekend(**kwargs)

    else:
        # is this a weekday or a weekend?
        if population.date is None:
            # have to guess from the day - assume day zero was a Monday
            day = population.day % 7
            # 0-4 is a weekday, 5 and 6 are weekend
            is_weekend = (day >= 5)
        else:
            day = population.date.weekday()
            # 0-4 is a weekday, 5 and 6 are weekend
            is_weekend = (day >= 5)

        if is_weekend:
            funcs = iterate_core(**kwargs) + iterate_weekend(**kwargs)
        else:
            funcs = iterate_core(**kwargs) + iterate_weekday(**kwargs)

    return funcs




__all__ = ["extract_default"]


def extract_default(nthreads: int = 1, setup=False, **kwargs):
    """This returns the default list of 'output_XXX' functions that
       are called in sequence for each iteration of the model run.
       These functions are used to output data to files for
       future processing

       Parameters
       ----------
       nthreads: int
         The number of threads that will be used for each function.
         If this is 1, then the serial versions of the functions will
         be returned, else the parallel (OpenMP) versions will be
         returned
       setup: bool
         Whether or not to return the functions used to setup the
         space and output files for the output_XXX functions returned by
         this iterator. This is called once at the start of a run
         to return the functions that must be called to setup the
         model

       Returns
       -------
       funcs: List[function]
         The list of functions that ```extract_data``` will call in sequence
    """

    if setup:
        from ._output_default import setup_output_default
        funcs = [setup_output_default]

    elif nthreads is None or nthreads == 1:
        from ._output_default import output_default
        funcs = [output_default]

    else:
        from ._output_default import output_default_omp
        funcs = [output_default_omp]

    return funcs




from typing import List
from itertools import chain
import copy

__all__ = ["string_to_ints"]


def rangeString(commaString):
    """Convert the passed comma-separated string of ints (including ranges)
       into a list of integers. Thanks for the great answers on
       stackoverflow
       https://stackoverflow.com/questions/6405208/how-to-convert-numeric-string-ranges-to-a-list-in-python

       This answer was provided by ninjagecko - Cheers :-)
    """
    def hyphenRange(hyphenString):
        hyphenString = hyphenString.strip()

        if len(hyphenString) == 0:
            return []

        try:
            x = [int(x) for x in hyphenString.split('-')]
            return range(x[0], x[-1]+1)
        except ValueError:
            # this can happen if it is a negative number
            if len(hyphenString) == 0:
                return []
            else:
                return [int(hyphenString)]

    return chain(*[hyphenRange(r) for r in commaString.split(',')])


def string_to_ints(string: str = None,
                   strings: List[str] = None):
    """Convert the passed string (or strings) containing
       integers (or ranges of integers) into a single sorted
       list of integers where no value is repeated
    """
    if string is not None:
        if strings is None:
            strings = []
        else:
            if len(strings) > 0:
                strings = copy.copy(strings)

        if isinstance(string, list):
            strings += string
        else:
            strings.append(string)

    ints = {}

    for string in strings:
        for i in rangeString(string):
            ints[int(i)] = 1

    ints = list(ints.keys())
    ints.sort()

    return ints




from .._network import Network
from .._outputfiles import OutputFiles
from ._workspace import Workspace
from .._population import Population, Populations
from ._profiler import Profiler, NullProfiler
from ._iterate import iterate
from ..iterators._iterate_default import iterate_default
from ..extractors._extract_default import extract_default
from ._clear_all_infections import clear_all_infections
from ._extract_data import extract_data

__all__ = ["run_model"]


def run_model(network: Network,
              infections, play_infections,
              rngs, s: int,
              output_dir: OutputFiles,
              population: Population = Population(initial=57104043),
              nsteps: int = None,
              profile: bool = True,
              profiler: Profiler = None,
              nthreads: int = None,
              iterator=None,
              extractor=None):
    """Actually run the model... Real work happens here. The model
       will run until completion or until 'nsteps' have been
       completed (whichever happens first)

        Parameters
        ----------
        network: Network
            The network on which to run the model
        infections: list
            The space used to record the infections
        play_infections: list
            The space used to record the play infectionss
        rngs: list
            The list of random number generators to use, one per thread
        population: Population
            The initial population at the start of the model outbreak.
            This is also used to set the date and day of the start of
            the model outbreak
        seed: int
            The random number seed used for this model run. If this is
            None then a very random random number seed will be used
        output_dir: OutputFiles
            The directory to write all of the output into
        nsteps: int
            The maximum number of steps to run in the outbreak. If None
            then run until the outbreak has finished
        profile: bool
            Whether or not to profile the model run and print out the
            results
        profiler: Profiler
            The profiler to use to profile - a new one is created if
            one isn't passed
        s: int
            Index of the seeding parameter to use
        nthreads: int
            Number of threads over which to parallelise this model run
        iterator: function
            Function that will be used to dynamically get the functions
            that will be used at each iteration to advance the
            model. Any additional files or parameters needed by these
            functions should be included in the `network.params` object.
        extractor: function
            Function that will be used to dynamically get the functions
            that will be used at each iteration to extract data from
            the model run

        Returns
        -------
        trajectory: Populations
            The trajectory of the population for every day of the model run
    """
    if iterator is None:
        iterator = iterate_default
    elif isinstance(iterator, str):
        from ..iterators._iterate_custom import build_custom_iterator
        iterator = build_custom_iterator(iterator, __name__)

    if extractor is None:
        extractor = extract_default
    elif isinstance(iterator, str):
        from ..extractors._extract_custom import build_custom_extractor
        extractor = build_custom_extractor(extractor, __name__)

    if profile:
        if profiler:
            p = profiler
        else:
            p = Profiler()
    else:
        p = NullProfiler()

    p = p.start("run_model")

    params = network.params

    if params is None:
        return population

    from copy import deepcopy
    population = deepcopy(population)

    # create space to hold the population trajectory
    trajectory = Populations()

    p = p.start("clear_all_infections")
    clear_all_infections(infections=infections,
                         play_infections=play_infections,
                         nthreads=nthreads)
    p = p.stop()

    # get and call all of the functions that need to be called to set
    # up the model run
    p = p.start("setup_funcs")
    setup_funcs = iterator(nthreads=nthreads, setup=True)

    for setup_func in setup_funcs:
        setup_func(network=network, population=population,
                   infections=infections, play_infections=play_infections,
                   rngs=rngs, profiler=p, nthreads=nthreads)

    setup_funcs = extractor(nthreads=nthreads, setup=True)

    for setup_func in setup_funcs:
        setup_func(network=network, population=population,
                   output_dir=output_dir, profiler=p, nthreads=nthreads)
    p = p.stop()

    # create a workspace that is used as part of extract_data to
    # provide a scratch-pad while extracting data from the model
    workspace = Workspace(network=network)

    # Now get the population and network data for the first day of the
    # model ("day zero", unless a future day has been set by the user)
    p = p.start("extract_data")
    extract_data(network=network, population=population, workspace=workspace,
                 output_dir=output_dir, infections=infections,
                 play_infections=play_infections,
                 rngs=rngs, get_output_functions=extractor,
                 nthreads=nthreads, profiler=p)
    p = p.stop()

    infecteds = population.infecteds

    # save the initial population
    trajectory.append(population)

    p = p.start("run_model_loop")
    iteration_count = 0

    # keep looping until the outbreak is over or until we have completed
    # at least 5 loop iterations
    while (infecteds != 0) or (iteration_count < 5):
        if profile:
            p2 = Profiler()
        else:
            p2 = NullProfiler()

        p2 = p2.start(f"timing for day {population.day}")

        iterate(network=network, population=population,
                infections=infections, play_infections=play_infections,
                rngs=rngs, get_advance_functions=iterator,
                nthreads=nthreads, profiler=p2)

        print(f"\n {population.day} {infecteds}")

        p2 = p2.start("extract_data")
        extract_data(network=network, population=population,
                     workspace=workspace, output_dir=output_dir,
                     infections=infections, play_infections=play_infections,
                     rngs=rngs, get_output_functions=extractor,
                     nthreads=nthreads, profiler=p2)
        p2 = p2.stop()

        infecteds = population.infecteds

        iteration_count += 1
        population.day += 1

        if population.date:
            from datetime import timedelta
            population.date += timedelta(days=1)

        if nsteps is not None:
            if iteration_count >= nsteps:
                trajectory.append(population)
                print(f"Exiting model run early at nsteps = {nsteps}")
                break

        p2 = p2.stop()

        if not p2.is_null():
            print(f"\n{p2}\n")

        # save the population trajectory
        trajectory.append(population)

    # end of while loop
    p = p.stop()
    p.stop()

    if not p.is_null():
        print(f"\nOVERALL MODEL TIMING\n{p}")

    print(f"Infection died ... Ending on day {population.day}")

    return trajectory




from dataclasses import dataclass

__all__ = ["Node"]


@dataclass
class Node:
    """This class represents an electoral ward (node) in the network"""

    #: The node's label (int). This is the index of the node, which must
    #: run from 1 up to len(nodes). A value of -1 implies a null node
    label: int = None

    #: The index (int) of the first link for this node in the
    #: network Links list
    begin_to: int = None

    #: The number (int) of links for this node in the Links list
    end_to: int = None

    self_w: int = None

    #: The index (int) of the first link for this node in the network
    #: play Links list (play)
    begin_p: int = None

    #: The number (int) of links for this node in the Links play list
    end_p: int = None
    self_p: int = None

    #: The index (int) of the first link for this node in the network
    #: weekend Links list
    begin_we: int = None
    end_we: int = None
    self_we: int = None

    day_foi: float = 0.0        # numerator only
    night_foi: float = 0.0      # numerator only
    weekend_foi: float = 0.0    # numerator only

    play_suscept: float = 0.0
    save_play_suscept: float = 0.0

    denominator_n: float = 0.0  #Denominator only
    denominator_d: float = 0.0  # Maybe won't need

    denominator_p: float = 0.0
    denominator_pd: float = 0.0

    #: x coordinate of the ward (location)
    x: float = 0.0

    #: y coordinate of the ward (location)
    y: float = 0.0

    b: float = 0.0

    #: the name of the ward, e.g. 'Bristol'
    id: str = None
    vacid: int = None




from dataclasses import dataclass as _dataclass
from typing import List as _List
from copy import deepcopy as _deepcopy
from datetime import date as _date

__all__ = ["Population", "Populations"]


@_dataclass
class Population:
    """This class holds information about the progress of the
       disease through the population
    """
    #: The initial population loaded into the model
    initial: int = 0
    #: The number of members who could be infected
    susceptibles: int = 0
    #: The number of latent infections
    latent: int = 0
    #: The total number of infections
    total: int = 0
    #: The total number who are removed from the outbreak,
    #: either because they have recovered, or are otherwise
    #: no longer able to be infected
    recovereds: int = 0
    #: The number infected in all wards
    n_inf_wards: int = 0
    #: The day in the outbreak of this record (e.g. day 0, day 10 etc.)
    day: int = 0
    #: The date in the outbreak of this record
    date: _date = None

    @property
    def population(self) -> int:
        """The total population in all wards"""
        return self.susceptibles + self.total + self.recovereds

    @property
    def infecteds(self) -> int:
        """The number who are infected across all wards"""
        return self.total + self.latent

    def __str__(self):
        s = f"DAY: {self.day} " \
            f"S: {self.susceptibles}    " \
            f"E: {self.latent}    " \
            f"I: {self.total}    " \
            f"R: {self.recovereds}    " \
            f"IW: {self.n_inf_wards}   " \
            f"TOTAL POPULATION {self.population}"

        if self.date:
            return f"{self.date.isoformat()}: {s}"
        else:
            return s


@_dataclass
class Populations:
    """This class holds the trajectory of Population objects recorded
       for every step (day) of a model outbreak
    """
    #: The trajectory of Population objects
    _trajectory: _List[Population] = None

    def __str__(self):
        if len(self) == 0:
            return "Populations:empty"
        else:
            return f"Latest: {self._trajectory[-1]}"

    def __getitem__(self, i: int):
        """Return the ith Population in the trajectory"""
        if self._trajectory is None:
            raise IndexError("No trajectory data collected")
        else:
            return self._trajectory[i]

    def __len__(self):
        if self._trajectory is None:
            return 0
        else:
            return len(self._trajectory)

    def append(self, population: Population):
        """Append the next step in the trajectory.

           Parameters
           ----------
           population: Population
             The population to append to this list
        """
        if not isinstance(population, Population):
            raise TypeError("Only Population objects should be recorded!")

        if self._trajectory is None:
            self._trajectory = []

        self._trajectory.append(_deepcopy(population))




from .._network import Network
from .._parameters import Parameters
from .._outputfiles import OutputFiles

import os
import sys
from contextlib import contextmanager

__all__ = ["run_worker", "prepare_worker"]

global_network = None


@contextmanager
def silence_output():
    """Nice way to silence stdout and stderr - thanks to
       Emil Stenström in
       https://stackoverflow.com/questions/6735917/redirecting-stdout-to-nothing-in-python
    """
    new_out = open(os.devnull, "w")
    old_out = sys.stdout
    sys.stdout = new_out

    new_err = open(os.devnull, "w")
    old_err = sys.stderr
    sys.stderr = new_err

    try:
        yield new_out
    finally:
        sys.stdout = old_out
        sys.stderr = old_err


def prepare_worker(params: Parameters):
    """Prepare a worker to receive work to run a model using the passed
       parameters. This will build the network specified by the
       parameters and will store it in global memory ready to
       be used for a model run. Note that these are
       silent, printing nothing to stdout or stderr
    """
    # switch off printing to stdout and stderr
    with silence_output():
        global global_network

        if global_network is None:
            global_network = Network.build(params=params,
                                           calculate_distances=True,
                                           profile=False)

        else:
            global_network.update(params)


def run_worker(arguments):
    """Ask the worker to run a model using the passed variables and
       options. This will write to options['output_dir'] and will
       also return the population object that contains the final
       population data.

       WARNING - the iterator and extractor arguments rely on the
       workers starting in the same directory as the main process,
       so that they can load the same python files (if the user
       is using a custom iterator or extractor)
    """
    params = arguments["params"]
    options = arguments["options"]

    # first, build and prepare the network
    prepare_worker(params)

    # next, run the job, writing to output
    from ._runner import redirect_output

    outdir = options["output_dir"]
    auto_bzip = options["auto_bzip"]
    del options["auto_bzip"]

    # if the user wanted to remove this directory then they would
    # have done so in the main process - no need to check again
    with OutputFiles(outdir, check_empty=False, force_empty=False,
                     prompt=None, auto_bzip=auto_bzip) as output_dir:
        options["output_dir"] = output_dir

        with redirect_output(output_dir.get_path()):
            global global_network
            output = global_network.run(**options)

            return output





from .._network import Network
from ._array import create_int_array

__all__ = ["initialise_infections",
           "initialise_play_infections"]


def initialise_infections(network: Network):
    """Initialise the data structure used to store the infections"""

    params = network.params

    if params is None:
        return None

    disease = params.disease_params

    n = disease.N_INF_CLASSES()

    infections = []

    # 'infections' holds all of the infections recorded for every
    # single link in network.to_links. 1-indexing is used throughout
    # the code, hence why we size for n + 1
    for _ in range(0, n):
        infections.append(create_int_array(network.nlinks + 1))

    return infections


def initialise_play_infections(network: Network):
    """Initialise the space used to store the play infections"""

    params = network.params

    if params is None:
        return None

    disease = params.disease_params

    n = disease.N_INF_CLASSES()

    infections = []

    # the 'play_infections' holds the infections that occur in
    # each ward (node) according to the 'play' rules. 1-indexing
    # is used throughout the code, hence why we size for n + 1
    for _ in range(0, n):
        infections.append(create_int_array(network.nnodes+1))

    return infections




def parse_args():
    """Parse all of the command line arguments"""
    import argparse
    import sys

    parser = argparse.ArgumentParser(
                    description="MetaWards simple plotting program - see "
                                "https://github.com/metawards/metawards "
                                "for more information",
                    prog="metawards-plot")

    parser.add_argument('--version', action="store_true",
                        default=None,
                        help="Print the version information about "
                             "metawards-plot")

    parser.add_argument("-i", "--input", type=str,
                        help="Full path to the 'results.csv.bz2' file "
                             "that you want to quickly plot.")

    parser.add_argument("-a", "--animate", type=str, nargs="*",
                        help="Full path to the set of plots that you "
                             "want to animate together into an animated "
                             "gif")

    parser.add_argument("-o", "--output", type=str, default=None,
                        help="Location for the output. By default "
                             "this is inferred from the input")

    parser.add_argument("--format", type=str, default="pdf",
                        help="Desired format of the graphs that are "
                             "produced. Good formats are 'pdf' for "
                             "publication graphics and 'png' or 'jpeg' "
                             "for webpages")

    parser.add_argument("--dpi", type=int, default=300,
                        help="Resolution to use when creating bitmap "
                             "outputs, e.g. jpg, png etc.")

    parser.add_argument("--delay", type=int, default=500,
                        help="The delay in milliseconds between animation "
                             "frames if an animation is being produced")

    parser.add_argument("--ordering", type=str, default="fingerprint",
                        help="Ordering to use for the frames. This "
                             "can be fingerprint, filename or custom")

    args = parser.parse_args()

    if args.version:
        from metawards import get_version_string
        print(get_version_string())
        sys.exit(0)

    return (args, parser)


def cli():
    """Main function for the metawards-plot command line interface.
       This is a simple program that creates basic plots of the output
       data. The aim is to make visualisation of the data as quick
       and painless as possible.
    """
    args, parser = parse_args()

    if args.input:
        from metawards.analysis import save_summary_plots

        filenames = save_summary_plots(results=args.input,
                                       output_dir=args.output,
                                       format=args.format,
                                       dpi=args.dpi,
                                       verbose=True)

        print(f"Written graphs to {', '.join(filenames)}")

    elif args.animate is not None and len(args.animate) > 0:
        from metawards.analysis import animate_plots

        filename = animate_plots(plots=args.animate,
                                 output=args.output,
                                 delay=args.delay,
                                 ordering=args.ordering,
                                 verbose=True)

        print(f"Written animation to {filename}")
    else:
        import sys
        parser.print_help()
        sys.exit(0)


if __name__ == "__main__":
    cli()




__all__ = ["iterate_core"]


def iterate_core(nthreads: int = 1, setup=False, **kwargs):
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence at the beginning of each iteration of
       the model run. These are the core functions, so must

       Parameters
       ----------
       nthreads: int
         The number of threads that will be used for each function.
         If this is 1, then the serial versions of the functions will
         be returned, else the parallel (OpenMP) versions will be
         returned
       setup: bool
         Whether or not to return the functions used to setup the
         space and input for the advance_XXX functions returned by
         this iterator. This is called once at the start of a run
         to return the functions that must be called to setup the
         model.

       Returns
       -------
       funcs: List[function]
         The list of functions that ```iterate``` will call in sequence
    """

    if setup:
        # Return the functions needed to initialise this iterator
        from ._setup_imports import setup_seed_wards
        from ._advance_additional import setup_additional_seeds

        funcs = [setup_seed_wards,
                 setup_additional_seeds]

    elif nthreads is None or nthreads == 1:
        from ._advance_additional import advance_additional
        from ._advance_foi import advance_foi
        from ._advance_recovery import advance_recovery

        funcs = [advance_additional,
                 advance_foi,
                 advance_recovery]
    else:
        from ._advance_additional import advance_additional_omp
        from ._advance_foi import advance_foi_omp
        from ._advance_recovery import advance_recovery_omp

        funcs = [advance_additional_omp,
                 advance_foi_omp,
                 advance_recovery_omp]

    return funcs




from typing import List as _List

__all__ = ["animate_plots", "import_animate_modules"]


def import_animate_modules():
    """Import the python modules needed to animate plots. This
       imports Python Pillow

       Returns
       -------
       (Image, ImageDraw)
         The PIL.Image and PIL.ImageDraw modules
    """
    try:
        from PIL import Image, ImageDraw
    except ImportError:
        print("Could not import Python Pillow to create animated plot.")
        print("Please install using 'pip install Pillow>=6.2.1'")
        print("(or by running metawards-install --optional)")

    return (Image, ImageDraw)


def animate_plots(plots: _List[str], output: str,
                  delay: int = 500,
                  ordering: str = "fingerprint",
                  verbose = False):
    """Animate the plots contained in the filenames 'plots', writing
       the output to 'output'. Creates an animated gif of the plots
       and writes them to the file 'output'

       Parameters
       ----------
       plots: List[str]
         The list of files containing graphs to plot
       output: str
         The name of the file to write the animated gif to
       delay: int
         The delay in milliseconds between frames. Default is
         500 milliseconds (2 frames per second) which is reasonable
         for animated graphs
       ordering: str
         The ordering to use for the frames. This can be
         'fingerprint', 'filename' or 'custom'
       verbose: bool
         Whether to print out information to the screen during
         processing
    """

    if ordering == "fingerprint":
        from .._variableset import VariableSet

        if verbose:
            print("Arranging into fingerprint order...")

        fingerprints = {}

        for plot in plots:
            try:
                (values, repeat_idx) = VariableSet.extract_values(plot)
            except Exception as e:
                raise ValueError(
                    f"Could not extract a fingerprint from '{plot}'. "
                    f"Error was {e.__class__} {e}")

            if repeat_idx:
                values.append(repeat_idx)
            else:
                values.append(0)

            # now create a sortable string from these values
            key = " ".join([("%.5f" % x) for x in values])
            fingerprints[key] = plot

        keys = list(fingerprints.keys())
        keys.sort()

        plots = []
        for key in keys:
            plots.append(fingerprints[key])

    elif ordering == "filename":
        if verbose:
            print("Arranging into filename order...")

        plots.sort()

    elif ordering == "custom":
        if verbose:
            print("Using the user-supplied order...")

    else:
        print(f"Could not recognise ordering scheme {ordering}")
        raise ValueError(f"Unrecognised ordering scheme {ordering}")

    if output is None:
        import os
        common = os.path.commonprefix(plots)
        if not (common.endswith("_")):
            common += "_"

        output = f"{common}animate.gif"

    if verbose:
        print(f"Animating the below frame in this order, with a delay "
              f"between frames of {delay} ms.")

        for i, plot in enumerate(plots):
            print(f"{i+1}: {plot}")

        print(f"Output will be written to {output}")

    # open all of the images
    images = []

    (Image, ImageDraw) = import_animate_modules()

    for plot in plots:
        images.append(Image.open(plot))

    images[0].save(output, save_all=True, append_images=images[1:],
                   optimize=False, duration=delay, loop=0)

    # now try to optimize the gif
    try:
        from pygifsicle import optimize
        have_optimize = True
    except Exception:
        have_optimize = False

    if have_optimize:
        if verbose:
            print("Optimizing the resulting gif...")

        try:
            optimize(output)
        except Exception:
            if verbose:
                print("Optimisation failed - using unoptimized gif")

    return output



"""
.. currentmodule:: metawards.extractors

Functions
=========

.. autosummary::
    :toctree: generated/

    extract_default
    output_default
    output_default_omp
    setup_output_default

"""

from ._extract_default import *
from ._output_default import *




from libc.stdint cimport uintptr_t

cdef extern from "../ran_binomial/distributions.h":
    ctypedef struct binomial_rng

    binomial_rng* binomial_rng_alloc() nogil
    void binomial_rng_free (binomial_rng * r) nogil

    void seed_ran_binomial ( binomial_rng * rng, int seed) nogil
    unsigned int ran_binomial(binomial_rng * rng, double p, unsigned int n) nogil
    double binomial_rng_uniform(binomial_rng * rng) nogil


cdef inline uintptr_t _construct_binomial_rng() nogil:
    """Construct and return a pointer to a new gsl random number
       generator. This will return a uintptr_t that can be handled
       in Python. You must destroy this random number generator
       when you don't want it anymore, using the _delete_gsl_rng
       function
    """
    cdef binomial_rng* r
    r = binomial_rng_alloc()

    cdef uintptr_t p = <uintptr_t>r

    return p


cdef inline void _delete_binomial_rng(uintptr_t rng) nogil:
    """Delete the passed gsl random number generator.
       WARNING: It is not safe to use this pointer to
       generate any more random numbers!
    """
    cdef binomial_rng* r = _get_binomial_ptr(rng)
    binomial_rng_free(r)


cdef inline binomial_rng* _get_binomial_ptr(uintptr_t rng) nogil:
    """Get the underlying binomial_rng pointer from the uintptr_t"""
    cdef binomial_rng* r = <binomial_rng*>rng

    return r


cdef inline uintptr_t _seed_ran_binomial(uintptr_t rng,
                                         unsigned long seed) nogil:
    """Seed the passed random number generator"""
    cdef binomial_rng* r = _get_binomial_ptr(rng)
    seed_ran_binomial(r, seed)


cdef inline int _ran_binomial(binomial_rng *rng, double p, int n) nogil:
    """Generate a random number from the binomial distribution
       described by p and n
    """
    return ran_binomial(rng, p, n)

cdef inline double _ran_uniform(binomial_rng *rng) nogil:
    """Generate the next random-distributed uniform [0,1] random number"""
    return binomial_rng_uniform(rng)




from libc.math cimport exp

cdef inline double rate_to_prob(double rate) nogil:
    """Convert the return the probability associated with the passed
       infection rate
    """
    return rate - (0.5*rate*rate) if rate < 1e-6 else 1.0 - exp(-rate)



"""
.. currentmodule:: metawards

Classes
=======

.. autosummary::
    :toctree: generated/

    Disease
    InputFiles
    Link
    Links
    Network
    Node
    Nodes
    OutputFiles
    Parameters
    Population
    Populations
    VariableSet
    VariableSets

Functions
=========

.. autosummary::
    :toctree: generated/

    get_version_string

"""

__all__ = ["get_version_string"]

from ._parameters import *
from ._disease import *
from ._inputfiles import *
from ._node import *
from ._link import *
from ._population import *
from ._network import *
from ._outputfiles import *
from ._variableset import *

# import the pyx cython-compiled modules
from ._nodes import *
from ._links import *

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import utils
from . import iterators
from . import extractors
from . import app
from . import analysis


def get_version_string():
    """Return a version string for metawards which can be printed
       into a file or written out to the screen
    """
    from ._version import get_versions
    v = get_versions()

    header = f"metawards version {v['version']}"
    stars = "*" * len(header)

    def center(line, length=48):
        diff = length - len(line)
        if diff <= 0:
            return line
        else:
            left = int(diff/2) * " "
            return left + line

    newline = "\n"

    stars = center(stars)

    lines = []
    lines.append("")
    lines.append(stars)
    lines.append(center(header))
    lines.append(stars)
    lines.append("")

    lines.append(center(f"-- Source information --"))
    lines.append(center(f"repository: {v['repository']}"))
    lines.append(center(f"branch: {v['branch']}"))
    lines.append(center(f"revision: {v['full-revisionid']}"))
    lines.append(center(f"last modified: {v['date']}"))

    if v["dirty"]:
        lines.append("")
        lines.append("WARNING: This version has not been committed to git,")
        lines.append("WARNING: so you may not be able to recover the original")
        lines.append("WARNING: source code that was used to generate "
                     "this run!")

    lines.append("")
    lines.append(center(f"-- Additional information --"))
    lines.append(center(f"Visit https://metawards.github.io for more "
                        f"information"))
    lines.append(center(f"about metawards, its authors and its license"))
    lines.append("")

    return newline.join(lines)




__all__ = ["iterate_default", "iterator_needs_setup"]

from ._iterate_core import iterate_core
from ._iterate_weekday import iterate_weekday


def iterator_needs_setup(iterator):
    """Return whether or not the passed iterator function has
       a "setup" argument, and thus needs to be setup before
       it can be used
    """
    import inspect
    return "setup" in inspect.signature(iterator).parameters


def iterate_default(setup=False, **kwargs):
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence for each iteration of the model run.
       This is the default iterator. It models every day as though
       it is a working day.

       Parameters
       ----------
       setup: bool
         Whether or not to return the functions used to setup the
         space and input for the advance_XXX functions returned by
         this iterator. This is called once at the start of a run
         to return the functions that must be called to setup the
         model

       Returns
       -------
       funcs: List[function]
         The list of functions that ```iterate``` will call in sequence
    """

    kwargs["setup"] = setup

    if setup:
        # Return the functions needed to initialise this iterator
        funcs = iterate_core(**kwargs)

        # Does 'iterate_weekday' need to be setup?
        if iterator_needs_setup(iterate_weekday):
            funcs += iterate_weekday(**kwargs)

    else:
        funcs = iterate_core(**kwargs) + iterate_weekday(**kwargs)

    return funcs




from array import array

__all__ = ["guess_num_threads_and_procs",
           "get_available_num_threads",
           "create_thread_generators"]


def guess_num_threads_and_procs(njobs: int,
                                nthreads: int = None,
                                nprocs: int = None,
                                ncores: int = None,
                                parallel_scheme: str = None):
    """Guess the number of processes and threads per process
       to use to make most-efficient processing of
       'njobs' jobs.

       Parameters
       ----------
       njobs: int
         The number of jobs (model runs) that must be performed
       nthreads: int
         The number of threads requested - if None then this this will be
         guessed
       nprocs: int
         The number of processes requested - if None then this will be
         guessed
       ncores: int
         The number of available cores on each node (or this computer).
         This will be obtained from the OS if it is not supplied
       parallel_scheme: str
         The parallelisation scheme that is being used to parallelise
         over processes

       Returns
       -------
       (nthreads, nprocs): Tuple[int, int]
         Tuple of the best-guessed 'nthreads' and 'nprocs' values to use
    """
    if nthreads is None:
        nthreads = 0

    if nprocs is None:
        nprocs = 0

    if njobs is None or njobs < 0:
        return (1, 1)

    if parallel_scheme is None:
        parallel_scheme = "multiprocessing"

    if nprocs < 1 and parallel_scheme != "multiprocessing":
        # we have been told the number of processes to use by
        # the scoop or mpi4py scheduler
        from metawards.utils import get_number_of_processes
        nprocs = get_number_of_processes(parallel_scheme, nprocs)

    if ncores is None:
        ncores = get_available_num_threads()

    if nthreads < 1 and nprocs < 1:
        # we will need to calculate the best values of nprocs and
        # nthreads to use to divide the work most efficiently. The
        # goal should be to run as many model runs in parallel as
        # possible

        if njobs >= 0.8 * ncores:   # not worth going parallel for <80%
            nthreads = 1
            nprocs = ncores
        else:
            import math
            nthreads = int(math.floor(ncores/njobs))
            if nthreads < 1:
                nthreads = 1
            nprocs = int(math.floor(ncores/nthreads))
            if nprocs < 1:
                nprocs = 1
    elif nthreads < 1:
        import math
        nthreads = int(math.floor(ncores/nprocs))

        if nthreads < 1:
            nthreads = 1
    elif nprocs < 1:
        import math
        max_nprocs = int(math.floor(ncores/nthreads))
        if max_nprocs < njobs:
            nprocs = max_nprocs
        else:
            nprocs = max_nprocs

    return (nthreads, nprocs)


def get_available_num_threads():
    """Return the maximum number of threads that are recommended
       for this computer (the OMP_NUM_THREADS value)
    """
    import os

    # try OMP_NUM_THREADS as this is the accepted way to
    # override the number in a queueing system
    omp_num_threads = os.getenv("OMP_NUM_THREADS", None)

    if omp_num_threads is not None:
        try:
            return int(omp_num_threads)
        except Exception:
            pass

    # ok, get this from 'os'
    return os.cpu_count()


def create_thread_generators(rng, nthreads):
    """Return a set of random number generators, one for each
       thread - these are seeded using the next 'nthreads'
       random numbers drawn from the passed generator

       If 'nthreads' is 1, 0 or None, then then just
       returns the passed 'rng'
    """
    rngs = []

    if nthreads is None or nthreads <= 1:
        rngs.append(rng)
    else:
        from ._ran_binomial import seed_ran_binomial, ran_int

        for i in range(0, nthreads):
            seed = ran_int(rng)
            print(f"Random seed for thread {i} equals {seed}")
            rngs.append(seed_ran_binomial(seed))

    # need to return these as an array so that they are more easily
    # accessible from the OpenMP loops - rng is a unsigned 64bit integer
    # as a uintptr_t - this best corresponds to unsigned long ("L")
    return array("L", rngs)




from dataclasses import dataclass as _dataclass
import os as _os
import pathlib as _pathlib

__all__ = ["InputFiles"]

_inputfiles = {}

_default_model_path = _os.path.join(_pathlib.Path.home(),
                                    "GitHub", "MetaWardsData")

_default_folder_name = "model_data"


@_dataclass
class InputFiles:
    """This class holds all of the input files that must be loaded
       from METAWARDSDATA to construct the network of wards
       and links between them

       Load using the InputFiles.load function e.g.

       Examples
       --------
       >>> infiles = InputFiles.load("2011Data")
       >>> print(infiles)
       Model 2011Data version March 29 2020
       repository: https://github.com/metawards/MetaWardsData
       repository_branch: master
       repository_version: 0.2.0
       etc.
    """
    #: File from which to read the work matrix of links
    work: str = None
    #: File from which to read the play matrix of links
    play: str = None
    #: File from which to read all of the ward names
    identifier: str = None
    #: File from which to read all of the secondary
    #: ward IDs, (Communities, Counties, Districts, UA's etc)
    identifier2: str = None
    #: File from which to read the weekend matrix of links
    weekend: str = None
    #: File from which to read the size of the population in the play pile
    play_size: str = None
    #: File from which to read the positions (locations) of the wards
    #: (the centre of the bounding boxes)
    position: str = None
    #: File from which to read the values to seed the wards
    seed: str = None
    #: File from which to read the list of nodes to track
    nodes_to_track: str = None
    #: UV file
    uv: str = None

    _filename: str = None            # Full path of the description.json file
    _model_name: str = None          # Name of the model
    _model_path: str = None          # Directory containing the input files
    _model_version: str = None       # Version loaded from the data
    _authors: str = None             # Author(s) of this data
    _contacts: str = None            # Contact(s) for this data
    _references: str = None          # References for this data
    _repository: str = None          # GitHub repository for this data
    _repository_version: str = None  # Git version of the data
    _repository_branch: str = None   # Git branch for the data

    def model_name(self):
        """Return the name of this model"""
        return self._model_name

    def model_path(self):
        """Return the path to the directory containing this model"""
        return self._model_path

    def model_version(self):
        """Return the version of the data in this model"""
        return self._model_version

    def __str__(self):
        return f"Model {self._model_name} version {self._model_version}\n" \
               f"loaded from {self._filename}\n" \
               f"root directory {self._model_path}\n" \
               f"author(s): {self._authors}\n" \
               f"contact(s): {self._contacts}\n" \
               f"references(s): {self._references}\n" \
               f"repository: {self._repository}\n" \
               f"repository_branch: {self._repository_branch}\n" \
               f"repository_version: {self._repository_version}\n\n" \
               f"work = {self.work}\n" \
               f"play = {self.play}\n" \
               f"identifier = {self.identifier}\n" \
               f"identifier2 = {self.identifier2}\n" \
               f"weekend = {self.weekend}\n" \
               f"play_size = {self.play_size}\n" \
               f"position = {self.position}\n" \
               f"seed = {self.seed}\n" \
               f"nodes_to_track = {self.nodes_to_track}\n\n"

    def _localise(self):
        """Localise the filenames in this input files set. This will
           prepend model_path/model to every filename and will also
           double-check that all files exist and are readable
        """
        members = [attr for attr in dir(self)
                   if not callable(getattr(self, attr))
                   and not attr.startswith("_")]

        for member in members:
            filename = getattr(self, member)
            if filename:
                filename = _os.path.join(self._model_path, filename)

                if not (_os.path.exists(filename) and
                        _os.path.isfile(filename)):
                    raise FileNotFoundError(
                            f"Cannot find input file {member} = {filename}")

                setattr(self, member, filename)

    @staticmethod
    def load(model: str = "2011Data",
             repository: str = _default_model_path,
             folder: str = _default_folder_name,
             description: str = "description.json",
             filename: str = None):
        """Load the parameters associated with the passed model.
           This will look for the parameters specified in
           the json file called f"{repository}/{folder}/{model}/{description}"

           By default this will load the 2011Data parameters
           from $HOME/GitHub/model_data/2011Data/description.json

           Alternatively you can provide the full path to the
           description json file usng the 'filename' argument.
           All files within this description will be searched
           for using the directory that contains that file
           as a base

           Parameters
           ----------
           model: str
             The name of the model data to load. This is the name that
             will be searched for in the METAWARDSDATA model_data directory
           repository: str
             The location of the cloned METAWARDSDATA repository
           folder: str
             The name of the folder within the METAWARDSDATA repository
             that contains the model data
           filename: str
             The name of the file to load the model data from - this directly
             loads this file without searching through the METAWARDSDATA
             repository

           Returns
           -------
           input_files: InputFiles
             The constructed and validated set of input files
        """
        repository_version = None
        repository_branch = None

        if filename is None:
            if repository is None:
                repository = _os.getenv("METAWARDSDATA")
                if repository is None:
                    repository = _default_model_path

            filename = _os.path.join(repository, folder,
                                     model, description)

            from ._parameters import get_repository_version
            v = get_repository_version(repository)
            repository = v["repository"]
            repository_version = v["version"]
            repository_branch = v["branch"]

        json_file = filename
        model_path = _os.path.dirname(filename)

        try:
            with open(json_file, "r") as FILE:
                import json
                files = json.load(FILE)

        except Exception as e:
            print(f"Could not find the model file {json_file}")
            print(f"Either it does not exist of was corrupted.")
            print(f"Error was {e.__class__} {e}")
            print(f"To download the model data type the command:")
            print(f"  git clone https://github.com/metawards/MetaWardsData")
            print(f"and then re-run this function passing in the full")
            print(f"path to where you downloaded this directory")
            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")

        model = InputFiles(work=files["work"],
                           play=files["play"],
                           identifier=files["identifier"],
                           identifier2=files["identifier2"],
                           weekend=files["weekend"],
                           play_size=files["play_size"],
                           position=files["position"],
                           seed=files["seed"],
                           nodes_to_track=files["nodes_to_track"],
                           uv=files["uv"],
                           _filename=json_file,
                           _model_path=model_path,
                           _model_name=files["name"],
                           _model_version=files["version"],
                           _references=files["reference(s)"],
                           _authors=files["author(s)"],
                           _contacts=files["contact(s)"],
                           _repository=repository,
                           _repository_version=repository_version,
                           _repository_branch=repository_branch)

        model._localise()

        return model



#!/bin/env python3
"""
The metawards command line program.

Usage:
    To get the help for this program and the list of all of the
    arguments (with defaults) use;

    metawards --help
"""

__all__ = ["get_parallel_scheme", "parse_args", "get_hostfile",
           "get_cores_per_node", "get_threads_per_task",
           "scoop_supervisor", "mpi_supervisor",
           "cli"]


def get_parallel_scheme():
    """This function tries to work out which of the different parallel
       methods we should use to distribute work between multiple processes.

       Detected schemes are "mpi4py", "scoop", and if none of these
       are found, then "multiprocessing"
    """
    import sys

    if "mpi4py" in sys.modules:
        return "mpi4py"
    elif "scoop" in sys.modules:
        return "scoop"
    else:
        return "multiprocessing"


def parse_args():
    """Parse all of the command line arguments"""
    import argparse
    import sys

    parser = argparse.ArgumentParser(
                    description="MetaWards epidemic modelling - see "
                                "https://github.com/metawards/metawards "
                                "for more information",
                    prog="metawards")

    parser.add_argument('--version', action="store_true",
                        default=None,
                        help="Print the version information about metawards")

    parser.add_argument('-i', '--input', type=str,
                        help="Input file for the simulation that specifies "
                             "the adjustable parameters to change for each "
                             "run of the model. You must supply some "
                             "input to run a model!")

    parser.add_argument('-l', '--line', type=str, default=None, nargs="*",
                        help="Line number (or line numbers) containing the "
                             "values of adjustable parameters to run for this "
                             "run (or runs) of the model If this isn't "
                             "specified then model runs will be performed "
                             "for adjustable parameters given on all lines "
                             "from the input file. You can specify many "
                             "numbers, and ranges are also accepted, e.g. "
                             "'-l 5 6-10 12,13,14' etc. Note that the line "
                             "numbers are 0-indexed, e.g. the first line of "
                             "the file is line 0. Ranges are inclusive, "
                             "so 3-5 is the same as 3 4 5")

    parser.add_argument("-r", '--repeats', type=int, default=None,
                        help="The number of repeat runs of the model to "
                             "perform for each value of the adjustable "
                             "parameters. By default only a single "
                             "run will be performed for each set of "
                             "adjustable parameters")

    parser.add_argument('-s', '--seed', type=int, default=None,
                        help="Random number seed for this run "
                             "(default is to use a random seed)")

    parser.add_argument('-a', '--additional', type=str, default=None,
                        nargs="*",
                        help="File (or files) containing additional "
                             "seed of outbreak for the model. These are "
                             "used to seed additional infections on "
                             "specific days at different locations "
                             "during a model run")

    parser.add_argument('-o', '--output', type=str, default="output",
                        help="Path to the directory in which to place all "
                             "output files (default 'output'). This "
                             "directory will be subdivided if multiple "
                             "adjustable parameter sets or repeats "
                             "are used.")

    parser.add_argument('-u', '--UV', type=float, default=1.0,
                        help="Value for the UV parameter for the model "
                             "(default is 1.0)")

    parser.add_argument('-d', '--disease', type=str, default=None,
                        help="Name of the disease to model "
                             "(default is 'ncov')")

    parser.add_argument('-I', '--input-data', type=str, default="2011Data",
                        help="Name of the input data set for the network "
                             "(default is '2011Data')")

    parser.add_argument('--start-date', type=str, default=None,
                        help="Date of the start of the model outbreak. "
                             "This accepts dates either is iso-format, "
                             "or fuzzy dates such as 'monday', 'tomorrow' "
                             "etc. This is used to work out which days "
                             "are weekends, or to make it easier to specify "
                             "time-based events.")

    parser.add_argument('--start-day', type=int, default=0,
                        help="The start day of the model outbreak. By "
                             "default the model outbreak starts on day "
                             "zero (0), with each step of the model "
                             "representing an additional day. Use this "
                             "to start from a later day, which may be "
                             "useful if you want to more quickly reach "
                             "time based events. Note that the passed "
                             "'--start-date' is always day 0, so day 10 "
                             "has a date which is 10 days after start-date")

    parser.add_argument('-p', '--parameters', type=str, default="march29",
                        help="Name of the input parameter set used to "
                             "control the simulation (default 'march29')")

    parser.add_argument('-R', '--repository', type=str, default=None,
                        help="Path to the MetaWardsData repository. If "
                             "unspecified this defaults to the value "
                             "in the environment variable METAWARDSDATA "
                             "or, if that isn't specified, to "
                             "$HOME/GitHub/MetaWardsData")

    parser.add_argument('-P', '--population', type=int, default=57104043,
                        help="Initial population (default 57104043)")

    parser.add_argument('-n', '--nsteps', type=int, default=None,
                        help="Maximum number of steps (days) to run for the "
                             "simulation. Each step represents one day in the "
                             "outbreak (default is to run until the "
                             "outbreak has finished).")

    parser.add_argument('--iterator', type=str, default=None,
                        help="Name of the iterator to use to advance the "
                             "outbreak at each step (day). For a full "
                             "explanation see the tutorial at "
                             "https://metawards.github.io")

    parser.add_argument("--extractor", type=str, default=None,
                        help="Name of the extractor to use to extract "
                             "information during a model run. For a full "
                             "explanation see the tutorial at "
                             "https://metawards.github.io")

    parser.add_argument('--nthreads', type=int, default=None,
                        help="Number of threads over which parallelise an "
                             "individual model run. The total number of "
                             "cores used by metawards will be "
                             "nprocesses x nthreads")

    parser.add_argument("--nprocs", type=int, default=None,
                        help="The number of processes over which to "
                             "parallelise the different model runs for "
                             "different adjustable parameter sets and "
                             "repeats. By default this will automatically "
                             "work out the number of processes based on "
                             "the way metawards is launched. Use this "
                             "option if you want to specify the number "
                             "of processes manually.")

    parser.add_argument('--hostfile', type=str, default=None,
                        help="The hostfile containing the names of the "
                             "compute nodes over which to run a parallel "
                             "job. If this is not set, the program will "
                             "attempt to automatically get this information "
                             "from the cluster queueing system. Use this "
                             "if the auto-detection fails")

    parser.add_argument('--cores-per-node', type=int, default=None,
                        help="Set the number of processor cores available "
                             "on each of the compute nodes in the cluster "
                             "that will be used to run the models "
                             "(if a cluster is used). If this is not "
                             "set then the program will attempt to "
                             "get this information from the cluster "
                             "queueing system. Use this option if the "
                             "auto-detection fails.")

    parser.add_argument('--auto-bzip', action="store_true",
                        default=None,
                        help="Automatically bz2 compress "
                             "all output files as they are written.")

    parser.add_argument('--no-auto-bzip', action="store_true",
                        default=None,
                        help="Do not automatically bz2 compress "
                             "all output files as they are written.")

    parser.add_argument('--force-overwrite-output', action="store_true",
                        default=False,
                        help="Whether or not to force overwriting of any "
                             "existing output. Using this option will "
                             "tell metawards that it is safe to delete "
                             "the contents of the output directory "
                             "specified in by '--output' if it already "
                             "exists. Dangerous as this can remove "
                             "existing output files")

    parser.add_argument('--profile', action="store_true",
                        default=None, help="Enable profiling of the code")

    parser.add_argument('--no-profile', action="store_true",
                        default=None, help="Disable profiling of the code")

    parser.add_argument('--mpi', action="store_true", default=None,
                        help="Force use of MPI to parallelise across runs")

    parser.add_argument('--scoop', action="store_true", default=None,
                        help="Force use of scoop to parallelise across runs")

    # this hidden option is used to tell the main process started using
    # mpi that it shouldn't try to become a supervisor
    parser.add_argument('--already-supervised', action="store_true",
                        default=None, help=argparse.SUPPRESS)

    args = parser.parse_args()

    if args.version:
        from metawards import get_version_string
        print(get_version_string())
        sys.exit(0)

    return (args, parser)


def get_hostfile(args):
    """Attempt to find the name of the hostfile used to specify the hosts
       to use in a cluster
    """
    if args.hostfile:
        return args.hostfile

    import os

    # PBS
    hostfile = os.getenv("PBS_NODEFILE")

    if hostfile:
        return hostfile

    # SLURM
    hostfile = os.getenv("SLURM_HOSTFILE")

    if hostfile:
        return hostfile

    return None


def get_cores_per_node(args):
    """Return the number of cores per node in the cluster"""
    if args.cores_per_node:
        return args.cores_per_node

    import os

    try:
        cores_per_node = int(os.getenv("METAWARDS_CORES_PER_NODE"))

        if cores_per_node > 0:
            return cores_per_node

    except Exception:
        pass

    raise ValueError("You must specify the number of cores per node "
                     "using --cores-per-node or by setting the "
                     "environment variable METAWARDS_CORES_PER_NODE")


def get_threads_per_task(args):
    if args.nthreads:
        return args.nthreads

    import os

    try:
        nthreads = int(os.getenv("METAWARDS_THREADS_PER_TASK"))

        if nthreads > 0:
            return nthreads

        nthreads = int(os.getenv("OMP_NUM_THREADS"))

        if nthreads > 0:
            return nthreads
    except Exception:
        pass

    raise ValueError("You must specify the number of threads per task "
                     "using --nthreads or by setting the "
                     "environment variables METAWARDS_THREADS_PER_TASK "
                     "or OMP_NUM_THREADS")


def scoop_supervisor(hostfile, args):
    """Function used by the scoop supervisor to get the information needed to
       form the scoop call to run a scoop version of the program
    """
    import os
    import sys
    print("RUNNING A SCOOP PROGRAM")

    outdir = args.output

    if not os.path.exists(outdir):
        os.mkdir(outdir)

    cores_per_node = get_cores_per_node(args)

    print(f"Will run jobs assuming {cores_per_node} cores per compute node")

    # based on the number of threads requested and the number of cores
    # per node, we can work out the number of scoop processes to start,
    # and can write a hostfile that will create the right layout
    nthreads = get_threads_per_task(args)

    print(f"Will use {nthreads} OpenMP threads per model run...")

    tasks_per_node = int(cores_per_node / nthreads)

    print(f"...meaning that the number of model runs per node will be "
          f"{tasks_per_node}")

    # Next, read the hostfile to get a unique list of hostnames
    hostnames = {}

    with open(hostfile, "r") as FILE:
        line = FILE.readline()
        while line:
            hostname = line.strip()
            hostnames[hostname] = 1
            line = FILE.readline()

    hostnames = list(hostnames.keys())
    hostnames.sort()

    print(f"Number of compute nodes equals {len(hostnames)}")
    print(", ".join(hostnames))

    # how many tasks can we perform in parallel?
    nprocs = tasks_per_node * len(hostnames)

    if args.nprocs:
        if nprocs != args.nprocs:
            print(f"WARNING: You are using a not-recommended number of "
                  f"processes {args.nprocs} for the cluster {nprocs}.")

        nprocs = args.nprocs

    print(f"Total number of parallel processes to run will be {nprocs}")
    print(f"Total number of cores in use will be {nprocs*nthreads}")

    # Now write a new hostfile that round-robins the MPI tasks over
    # the nodes for 'tasks_per_node' runs
    hostfile = os.path.join(outdir, "hostfile")
    print(f"Writing hostfile to {hostfile}")

    with open(hostfile, "w") as FILE:
        i = 0
        while i < nprocs:
            for hostname in hostnames:
                FILE.write(hostname + "\n")
                i += 1

                if i == nprocs:
                    break

    # now craft the scoop command that will use this hostfile to
    # run the job - remember to pass the option to stop the main process
    # attempting to become a supervisor itself...

    import subprocess
    import shlex

    pyexe = sys.executable
    script = os.path.abspath(sys.argv[0])
    args = " ".join(sys.argv[1:])

    # also need to tell the main program the number of processes
    # as it can't work it out itself
    cmd = f"{pyexe} -m scoop --hostfile {hostfile} -n {nprocs} " \
          f"{script} --already-supervised {args} --nprocs {nprocs}"

    print(f"Executing scoop job using '{cmd}'")

    try:
        args = shlex.split(cmd)
        subprocess.run(args).check_returncode()
    except Exception as e:
        print("ERROR: Something went wrong!")
        print(f"{e.__class__}: {e}")
        sys.exit(-1)

    print("Scoop processes completed successfully")


def mpi_supervisor(hostfile, args):
    """Function used by the MPI supervisor to get the information needed to
       form the mpiexec call to run an MPI version of the program
    """
    import os
    import sys
    print("RUNNING AN MPI PROGRAM")

    outdir = args.output

    if not os.path.exists(outdir):
        os.mkdir(outdir)

    cores_per_node = get_cores_per_node(args)

    print(f"Will run jobs assuming {cores_per_node} cores per compute node")

    # based on the number of threads requested and the number of cores
    # per node, we can work out the number of mpi processes to start,
    # and can write a hostfile that will create the right layout
    nthreads = get_threads_per_task(args)

    print(f"Will use {nthreads} OpenMP threads per model run...")

    tasks_per_node = int(cores_per_node / nthreads)

    print(f"...meaning that the number of model runs per node will be "
          f"{tasks_per_node}")

    # Next, read the hostfile to get a unique list of hostnames
    hostnames = {}

    with open(hostfile, "r") as FILE:
        line = FILE.readline()
        while line:
            hostname = line.strip()
            hostnames[hostname] = 1
            line = FILE.readline()

    hostnames = list(hostnames.keys())
    hostnames.sort()

    print(f"Number of compute nodes equals {len(hostnames)}")
    print(", ".join(hostnames))

    # how many tasks can we perform in parallel?
    nprocs = tasks_per_node * len(hostnames)

    if args.nprocs:
        if nprocs != args.nprocs:
            print(f"WARNING: You are using an unrecommended number of "
                  f"processes {args.nprocs} for the cluster {nprocs}.")

        nprocs = args.nprocs

    print(f"Total number of parallel processes to run will be {nprocs}")
    print(f"Total number of cores in use will be {nprocs*nthreads}")

    # Now write a new hostfile that round-robins the MPI tasks over
    # the nodes for 'tasks_per_node' runs
    hostfile = os.path.join(outdir, "hostfile")
    print(f"Writing hostfile to {hostfile}")

    with open(hostfile, "w") as FILE:
        i = 0
        while i < nprocs:
            for hostname in hostnames:
                FILE.write(hostname + "\n")
                i += 1

                if i == nprocs:
                    break

    # now craft the mpiexec command that will use this hostfile to
    # run the job - remember to pass the option to stop the main process
    # attempt to become a supervisor itself...
    mpiexec = os.getenv("MPIEXEC")

    if mpiexec is None:
        mpiexec = "mpiexec"

    # check for weird mpiexecs...
    import subprocess
    import shlex

    try:
        args = shlex.split(f"{mpiexec} -v")
        p = subprocess.run(args, stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
        v = p.stdout.decode("utf-8").strip()
        print(f"{mpiexec} -v => {v}")

        if v.find("HPE HMPT") != -1:
            raise ValueError(
                "metawards needs a more modern MPI library than HPE's, "
                "so please compile to another MPI and use that.")
    except Exception as e:
        print(f"[ERROR] {e.__class__} {e}")

    pyexe = sys.executable
    script = os.path.abspath(sys.argv[0])
    args = " ".join(sys.argv[1:])

    cmd = f"{mpiexec} -np {nprocs} -hostfile {hostfile} " \
          f"{pyexe} -m mpi4py {script} --already-supervised {args}"

    print(f"Executing MPI job using '{cmd}'")

    try:
        args = shlex.split(cmd)
        subprocess.run(args).check_returncode()
    except Exception as e:
        print("ERROR: Something went wrong!")
        print(f"{e.__class__}: {e}")
        sys.exit(-1)

    print("MPI processes completed successfully")


def cli():
    """Main function for the command line interface. This does one of three
       things:

       1. If this is the main process, then it parses the arguments and
          runs and manages the jobs

       2. If this is a worker process, then it starts up and waits for work

       3. If this is a supervisor process, then it query the job scheduling
          system for information about the compute nodes to use, and will then
          set up and run a manager (main) process that will use those
          nodes to run the jobs
    """

    # get the parallel scheme now before we import any other modules
    # so that it is clear if mpi4py or scoop (or another parallel module)
    # has been imported via the required "-m module" syntax
    parallel_scheme = get_parallel_scheme()

    if parallel_scheme == "mpi4py":
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        nprocs = comm.Get_size()
        rank = comm.Get_rank()

        if rank != 0:
            # this is a worker process, so should not do anything
            # more until it is given work in the pool
            print(f"Starting worker process {rank+1} of {nprocs-1}...")
            return
        else:
            print("Starting main process...")

    elif parallel_scheme == "scoop":
        print("STARTING SCOOP PROCESS")

    import sys

    args, parser = parse_args()

    if not args.already_supervised:
        hostfile = get_hostfile(args)
        if hostfile:
            # The user has asked to run a parallel job - this means that this
            # process is the parallel supervisor
            if args.mpi:
                mpi_supervisor(hostfile, args)
                return
            elif args.scoop:
                scoop_supervisor(hostfile, args)
                return

            # neither is preferred - if scoop is installed then use that
            try:
                import scoop
                have_scoop = True
            except Exception:
                have_scoop = False

            if have_scoop:
                scoop_supervisor(hostfile, args)
                return

            # do we have MPI?
            try:
                import mpi4py
                have_mpi4py = True
            except Exception:
                have_mpi4py = False

            if have_mpi4py:
                mpi_supervisor(hostfile, args)
                return

            # we don't have any other option, just keep going and
            # use multiprocessing - in this case we don't need a
            # supervisor and this is the main process

    # This is now the code for the main process

    # WE NEED ONE OF these listed options;
    should_run = False

    for arg in [args.input, args.repeats, args.disease, args.additional]:
        if arg is not None:
            should_run = True
            break

    if not should_run:
        parser.print_help(sys.stdout)
        sys.exit(0)

    if args.repeats is None:
        args.repeats = 1

    # import the parameters here to speed up the display of help
    from metawards import Parameters, Network, Population, get_version_string

    # print the version information first, so that there is enough
    # information to enable someone to reproduce this run
    print(get_version_string())

    # also print the full command line used for this job
    print(f"Command used to run this job:\n{' '.join(sys.argv)}\n")

    if args.input:
        # get the line numbers of the input file to read
        if args.line is None or len(args.line) == 0:
            linenums = None
            print(f"Using parameters from all lines of {args.input}")
        else:
            from metawards.utils import string_to_ints
            linenums = string_to_ints(args.line)

            if len(linenums) == 0:
                print(f"You cannot read no lines from {args.input}?")
                sys.exit(-1)
            elif len(linenums) == 1:
                print(f"Using parameters from line {linenums[0]} of "
                      f"{args.input}")
            else:
                print(f"Using parameters from lines {linenums} of "
                      f"{args.input}")

        from metawards import VariableSets, VariableSet
        variables = VariableSets.read(filename=args.input,
                                      line_numbers=linenums)
    else:
        from metawards import VariableSets, VariableSet
        # create a VariableSets with one null VariableSet
        variables = VariableSets()
        variables.append(VariableSet())

    nrepeats = args.repeats

    if nrepeats is None or nrepeats < 1:
        nrepeats = 1

    if nrepeats == 1:
        print("Performing a single run of each set of parameters")
    else:
        print(f"Performing {nrepeats} runs of each set of parameters")

    variables = variables.repeat(nrepeats)

    # working out the number of processes and threads...
    from metawards.utils import guess_num_threads_and_procs
    (nthreads, nprocs) = guess_num_threads_and_procs(
                                            njobs=len(variables),
                                            nthreads=args.nthreads,
                                            nprocs=args.nprocs,
                                            parallel_scheme=parallel_scheme)

    print(f"\nNumber of threads to use for each model run is {nthreads}")

    if nprocs > 1:
        print(f"Number of processes used to parallelise model "
              f"runs is {nprocs}")
        print(f"Parallelisation will be achieved using {parallel_scheme}")

    # sort out the random number seed
    seed = args.seed

    if seed is None:
        import random
        seed = random.randint(10000, 99999999)

    print(f"\nUsing random number seed {seed}")

    # get the starting day and date
    start_day = args.start_day

    if start_day < 0:
        raise ValueError(f"You cannot use a start day {start_day} that is "
                         f"less than zero!")

    start_date = None

    if args.start_date:
        try:
            from dateparser import parse
            start_date = parse(args.start_date).date()
        except Exception:
            pass

        if start_date is None:
            from datetime import date
            try:
                start_date = date.fromisoformat(args.start_date)
            except Exception as e:
                raise ValueError(
                        f"Cannot interpret a valid date from "
                        f"'{args.start_date}'. Error is "
                        f"{e.__class__} {e}")

    if start_date is None:
        from datetime import date
        start_date = date.today()

    print(f"\nDay zero is {start_date.strftime('%A %B %d %Y')}")

    if start_day != 0:
        from datetime import timedelta
        start_day_date = start_date + timedelta(days=start_day)
        print(f"Starting on day {start_day}, which is "
              f"{start_day_date.strftime('%A %B %d %Y')}")
    else:
        start_day_date = start_date

    # now find the MetaWardsData repository as this will be needed
    # for the repeat command line too
    (repository, repository_version) = Parameters.get_repository(
                                                        args.repository)

    print(f"\nUsing MetaWardsData at {repository}")
    print(f"This is cloned from {repository_version['repository']}")
    print(f"branch {repository_version['branch']}, version "
          f"{repository_version['version']}")

    if repository_version["is_dirty"]:
        print("## WARNING - this repository is dirty, meaning that the data")
        print("## WARNING - has not been committed to git. This may make ")
        print("## WARNING - this calculation very difficult to reproduce")

    # now work out the minimum command line needed to repeat this job
    args.seed = seed
    args.nprocs = nprocs
    args.nthreads = nthreads
    args.start_date = start_date.isoformat()
    args.repository = repository

    repeat_cmd = "metawards"

    for key, value in vars(args).items():
        if value is not None:
            k = key.replace("_", "-")

            if isinstance(value, bool):
                if value:
                    repeat_cmd += f" --{k}"
            elif isinstance(value, list):
                repeat_cmd += f" --{k}"
                for val in value:
                    v = str(val)
                    if " " in v:
                        repeat_cmd += f" '{v}''"
                    else:
                        repeat_cmd += f" {v}"
            else:
                v = str(value)
                if " " in v:
                    repeat_cmd += f" --{k} '{v}''"
                else:
                    repeat_cmd += f" --{k} {v}"

    t = "*** To repeat this job use the command ***"

    print("\n" + "*"*len(t))
    print(t)
    print("*"*len(t) + "\n")
    print(repeat_cmd + "\n")

    # load all of the parameters
    try:
        params = Parameters.load(parameters=args.parameters)
    except Exception as e:
        print(f"Unable to load parameter files. Make sure that you have "
              f"cloned the MetaWardsData repository and have set the "
              f"environment variable METAWARDSDATA to point to the "
              f"local directory containing the repository, e.g. the "
              f"default is $HOME/GitHub/MetaWardsData")
        raise e

    # should we profile the code? (default yes, as it doesn't cost anything)
    profile = args.profile

    if args.no_profile:
        profile = False
    elif profile is None:
        profile = False

    # load the disease and starting-point input files
    if args.disease:
        params.set_disease(args.disease)
    else:
        params.set_disease("ncov")

    params.set_input_files(args.input_data)

    # read the additional seeds
    if args.additional is None or len(args.additional) == 0:
        print("Not using any additional seeds...")
    else:
        for additional in args.additional:
            print(f"Loading additional seeds from {additional}")
            params.add_seeds(additional)

    # extra parameters that are set
    params.UV = args.UV

    # set these extra parameters to 0
    params.static_play_at_home = 0
    params.play_to_work = 0
    params.work_to_play = 0
    params.daily_imports = 0.0

    # the size of the starting population
    population = Population(initial=args.population,
                            date=start_day_date,
                            day=start_day)

    print("\nBuilding the network...")
    network = Network.build(params=params, calculate_distances=True,
                            profile=profile)

    print("\nRun the model...")
    from metawards import OutputFiles
    from metawards.utils import run_models

    outdir = args.output

    if outdir is None:
        outdir = "output"

    if args.force_overwrite_output:
        prompt = None
    else:
        prompt = input

    auto_bzip = True

    if args.auto_bzip:
        auto_bzip = True
    elif args.no_auto_bzip:
        auto_bzip = False

    if args.iterator:
        iterator = args.iterator
        # eventually I should get the filename and function from this,
        # and then convert it into an absolute path to make this safe
        # on a cluster. Then I should copy the file into the output to
        # make sure that the calculation is reproducible. Will do this
        # when I copy this work to extractor
    else:
        iterator = None

    if args.extractor:
        extractor = args.extractor
        # see above ;-)
    else:
        extractor = None

    with OutputFiles(outdir, force_empty=args.force_overwrite_output,
                     auto_bzip=auto_bzip, prompt=prompt) as output_dir:
        result = run_models(network=network, variables=variables,
                            population=population, nprocs=nprocs,
                            nthreads=nthreads, seed=seed,
                            nsteps=args.nsteps,
                            output_dir=output_dir,
                            iterator=iterator,
                            extractor=extractor,
                            profile=profile, parallel_scheme=parallel_scheme)

        if result is None or len(result) == 0:
            print("No output - end of run")
            return 0

        # write the result to a csv file that can be easily read by R or
        # pandas - this will be written compressed to save space
        RESULTS = output_dir.open("results.csv", auto_bzip=auto_bzip)
        print(f"\nWriting a summary of all results into the\n"
              f"csv file {output_dir.get_filename('results.csv')}.\n"
              f"You can use this to quickly look at statistics across\n"
              f"all runs using e.g. R or pandas")

        varnames = result[0][0].variable_names()

        if varnames is None or len(varnames) == 0:
            varnames = ""
        else:
            varnames = ",".join(varnames) + ","

        has_date = result[0][1][0].date

        if has_date:
            datestring = "date,"
        else:
            datestring = ""

        RESULTS.write(f"fingerprint,repeat,{varnames}"
                      f"day,{datestring}S,E,I,R,IW\n")
        for varset, trajectory in result:
            varvals = varset.variable_values()
            if varvals is None or len(varvals) == 0:
                varvals = ""
            else:
                varvals = ",".join(map(str, varvals)) + ","

            start = f"{varset.fingerprint()}," \
                    f"{varset.repeat_index()},{varvals}"

            for i, pop in enumerate(trajectory):
                if pop.date:
                    d = pop.date.isoformat() + ","
                else:
                    d = ""

                RESULTS.write(f"{start}{pop.day},{d}{pop.susceptibles},"
                              f"{pop.latent},{pop.total},"
                              f"{pop.recovereds},{pop.n_inf_wards}\n")

    print("End of the run")

    return 0


if __name__ == "__main__":
    cli()

else:
    # this is one of the worker processes - make sure that they
    # have imported metawards
    from metawards.utils import run_worker




from .._network import Network
from .._population import Population

from ._profiler import Profiler, NullProfiler

__all__ = ["iterate"]


def iterate(network: Network, population: Population,
            infections, play_infections,
            rngs, nthreads: int,
            get_advance_functions,
            profiler: Profiler = None):
    """Advance the infection by one day for the passed Network,
       acting on the passed Population.

       Parameters
       ----------
       network: Network
         The network in which the disease outbreak will be modelled
       population: Population
         The population experiencing the outbreak. This contains
         an overview of the current population, plus the day and
         date of the outbreak
       infections
         Space in which the 'work' (fixed) infections are recorded
       play_infections
         Space in which the 'play' (random) infections are recorded
       rngs
         List of the thread-safe random number generators (one per thread)
       nthreads: int
         The number of threads over which to parallelise the calculation
       get_advance_functions: function
         This is a function that should return the set of "advance_XXX"
         functions that will be applied as part of this iteration
       profiler: Profiler
         The profiler to use to profile this calculation. Pass "None"
         if you want to disable profiling
    """
    if profiler is None:
        profiler = NullProfiler()

    p = profiler.start("iterate")

    advance_functions = get_advance_functions(network=network,
                                              population=population,
                                              infections=infections,
                                              play_infections=play_infections,
                                              rngs=rngs, nthreads=nthreads,
                                              profiler=p)

    for advance_function in advance_functions:
        advance_function(network=network,
                         population=population,
                         infections=infections,
                         play_infections=play_infections,
                         rngs=rngs, nthreads=nthreads,
                         profiler=p)

    p.stop()




from dataclasses import dataclass as _dataclass
from typing import List as _List
import pathlib as _pathlib
import os as _os

__all__ = ["Disease"]

_default_disease_path = _os.path.join(_pathlib.Path.home(),
                                      "GitHub", "MetaWardsData")

_default_folder_name = "diseases"


@_dataclass
class Disease:
    """This class holds the parameters about a single disease

       A disease is characterised as a serious of stages, each
       with their own values of the beta, progress, too_ill_to_move
       and contrib_foi parameters. To load a disease use
       the Disease.load function, e.g.

       Examples
       --------
       >>> disease = Disease.load("ncov")
       >>> print(disease)
       Disease ncov
       repository: https://github.com/metawards/MetaWardsData
       repository_branch: master
       repository_version: 0.2.0
       beta = [0.0, 0.0, 0.95, 0.95, 0.0]
       progress = [1.0, 0.1923, 0.909091, 0.909091, 0.0]
       too_ill_to_move = [0.0, 0.0, 0.0, 0.0, 0.0]
       contrib_foi = [1.0, 1.0, 1.0, 1.0, 0.0]
    """
    #: Beta parameter for each stage of the disease
    beta: _List[float] = None
    #: Progress parameter for each stage of the disease
    progress: _List[float] = None
    #: TooIllToMove parameter for each stage of the disease
    too_ill_to_move: _List[float] = None
    #: Contribution to the Force of Infection (FOI) parameter for each
    #: stage of the disease
    contrib_foi: _List[float] = None

    _name: str = None
    _version: str = None
    _authors: str = None
    _contacts: str = None
    _references: str = None
    _filename: str = None
    _repository: str = None
    _repository_version: str = None
    _repository_branch: str = None

    def __str__(self):
        return f"Disease {self._name}\n" \
               f"loaded from {self._filename}\n" \
               f"version: {self._version}\n" \
               f"author(s): {self._authors}\n" \
               f"contact(s): {self._contacts}\n" \
               f"references(s): {self._references}\n" \
               f"repository: {self._repository}\n" \
               f"repository_branch: {self._repository_branch}\n" \
               f"repository_version: {self._repository_version}\n\n" \
               f"beta = {self.beta}\n" \
               f"progress = {self.progress}\n" \
               f"too_ill_to_move = {self.too_ill_to_move}\n" \
               f"contrib_foi = {self.contrib_foi}\n\n"

    def __eq__(self, other):
        return self.beta == other.beta and \
               self.progress == other.progress and \
               self.too_ill_to_move == other.too_ill_to_move and \
               self.contrib_foi == other.contrib_foi

    def __len__(self):
        if self.beta:
            return len(self.beta)
        else:
            return 0

    def N_INF_CLASSES(self):
        """Return the number of stages of the disease"""
        return len(self.beta)

    def _validate(self):
        """Check that the loaded parameters make sense"""
        try:
            n = len(self.beta)

            assert len(self.progress) == n
            assert len(self.too_ill_to_move) == n
            assert len(self.contrib_foi) == n
        except Exception as e:
            raise AssertionError(f"Data read for disease {self._name} "
                                 f"is corrupted! {e.__class__}: {e}")

    @staticmethod
    def load(disease: str = "ncov",
             repository: str = None,
             folder: str = _default_folder_name,
             filename: str = None):
        """Load the disease parameters for the specified disease.
           This will look for a file called f"{disease}.json"
           in the directory f"{repository}/{disease}/{disease}.ncon"

           By default this will load the ncov (SARS-Cov-2)
           parameters from
           $HOME/GitHub/model_data/2011Data/diseases/ncov.json

           Alternatively you can provide the full path to the
           json file via the "filename" argument

           Parameters
           ----------
           disease: str
             The name of the disease to load. This is the name that
             will be searched for in the METAWARDSDATA diseases directory
           repository: str
             The location of the cloned METAWARDSDATA repository
           folder: str
             The name of the folder within the METAWARDSDATA repository
             that contains the diseases
           filename: str
             The name of the file to load the disease from - this directly
             loads this file without searching through the METAWARDSDATA
             repository

           Returns
           -------
           disease: Disease
             The constructed and validated disease
        """
        repository_version = None
        repository_branch = None

        if filename is None:
            import os
            if os.path.exists(disease):
                filename = disease
            elif os.path.exists(f"{disease}.json"):
                filename = f"{disease}.json"

        if filename is None:
            if repository is None:
                repository = _os.getenv("METAWARDSDATA")
                if repository is None:
                    repository = _default_disease_path

            filename = _os.path.join(repository, folder,
                                     f"{disease}.json")

            from ._parameters import get_repository_version
            v = get_repository_version(repository)
            repository = v["repository"]
            repository_version = v["version"]
            repository_branch = v["branch"]

        json_file = filename

        try:
            with open(json_file, "r") as FILE:
                import json
                data = json.load(FILE)

        except Exception as e:
            print(f"Could not find the disease file {json_file}")
            print(f"Either it does not exist of was corrupted.")
            print(f"Error was {e.__class__} {e}")
            print(f"To download the disease data type the command:")
            print(f"  git clone https://github.com/metawards/MetaWardsData")
            print(f"and then re-run this function passing in the full")
            print(f"path to where you downloaded this directory")
            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")

        disease = Disease(beta=data["beta"],
                          progress=data["progress"],
                          too_ill_to_move=data["too_ill_to_move"],
                          contrib_foi=data["contrib_foi"],
                          _name=disease,
                          _authors=data["author(s)"],
                          _contacts=data["contact(s)"],
                          _references=data["reference(s)"],
                          _filename=json_file,
                          _repository=repository,
                          _repository_branch=repository_branch,
                          _repository_version=repository_version)

        disease._validate()

        return disease




__all__ = ["iterate_custom",
           "build_custom_iterator"]

from ._iterate_core import iterate_core
from ._iterate_default import iterator_needs_setup


def build_custom_iterator(custom_function, parent_name="__main__"):
    """Build and return a custom iterator from the passed
       function. This will wrap 'iterate_custom' around
       the function to double-check that the custom
       function is doing everything correctly

       Parameters
       ----------
       custom_function
         This can either be a function, which will be wrapped and
         returned, or it can be a string. If it is a string then
         we will attempt to locate or import the function associated
         with that string. The search order is;

         1. Is this 'metawards.iterators.custom_function'?
         2. Is this 'custom_function' that is already imported'?
         3. Is this a file name in the current path, if yes then
            find the function in that file (either the first function
            called 'iterateXXX' or the specified function if
            custom_function is in the form module::function)

        parent_name: str
          This should be the __name__ of the calling function, e.g.
          call this function as build_custom_iterator(func, __name__)

        Returns
        -------
        iterator
          The wrapped iterator that is suitable for using in the iterate
          function.
    """
    if isinstance(custom_function, str):
        print(f"Importing a custom iterator from {custom_function}")

        # we need to find the function
        import metawards.iterators

        # is it metawards.iterators.{custom_function}
        try:
            func = getattr(metawards.iterators, custom_function)
            return build_custom_iterator(func)
        except Exception:
            pass

        # do we have the function in the current namespace?
        import sys
        try:
            func = getattr(sys.modules[__name__], custom_function)
            return build_custom_iterator(func)
        except Exception:
            pass

        # how about the __name__ namespace of the caller
        try:
            func = getattr(sys.modules[parent_name], custom_function)
            return build_custom_iterator(func)
        except Exception:
            pass

        # how about the __main__ namespace (e.g. if this was loaded
        # in a script)
        try:
            func = getattr(sys.modules["__main__"], custom_function)
            return build_custom_iterator(func)
        except Exception:
            pass

        # can we import this function as a file - need to check that
        # the user hasn't written this as module::function
        if custom_function.find("::") != -1:
            parts = custom_function.split("::")
            func_name = parts[-1]
            func_module = "::".join(parts[0:-1])
        else:
            func_name = None
            func_module = custom_function

        try:
            import importlib
            module = importlib.import_module(func_module)
        except Exception:
            module = None

        if module is None:
            try:
                import importlib.util
                spec = importlib.util.spec_from_file_location(
                                                func_module,
                                                f"{func_module}.py")

                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                print(f"Loaded iterator from {func_module}.py")
            except Exception:
                module = None

        if module is None:
            # we cannot find the iterator
            print(f"Cannot find the iterator '{custom_function}'."
                  f"Please make sure this is spelled correctly and "
                  f"any python modules/files needed are in the "
                  f"PYTHONPATH or current directory")
            raise ImportError(f"Could not import the iterator "
                              f"'{custom_function}'")

        if func_name is None:
            # find the first function that starts with 'iterate'
            import inspect
            for name, value in inspect.getmembers(module):
                if name.startswith("iterate"):
                    if hasattr(value, "__call__"):
                        # this is a function
                        return build_custom_iterator(getattr(module, name))

            print(f"Could not find any function in the module "
                  f"{custom_function} that has a name that starts "
                  f"with 'iterate'. Please manually specify the "
                  f"name using the '{custom_function}::your_function syntax")

            raise ImportError(f"Could not import the iterator "
                              f"{custom_function}")

        else:
            if hasattr(module, func_name):
                return build_custom_iterator(getattr(module, func_name))

            print(f"Could not find the function {func_name} in the "
                  f"module {func_module}. Check that the spelling "
                  f"is correct and that the right version of the module "
                  f"is being loaded.")
            raise ImportError(f"Could not import the iterator "
                              f"{custom_function}")

    if not hasattr(custom_function, "__call__"):
        print(f"Cannot build an iterator for {custom_function} "
              f"as it is missing a __call__ function, i.e. it is "
              f"not a function.")
        raise ValueError(f"You can only build custom iterators for "
                         f"actual functions... {custom_function}")

    print(f"Building a custom iterator for {custom_function}")

    return lambda **kwargs: iterate_custom(custom_function=custom_function,
                                           **kwargs)


def iterate_custom(custom_function,
                   setup=False, **kwargs):
    """This returns the default list of 'advance_XXX' functions that
       are called in sequence for each iteration of the model run.
       This iterator provides a custom iterator that uses
       'custom_function' passed from the user. This iterator makes
       sure that 'setup' is called correctly and that the functions
       needed by 'iterate_core' are called first.

       Parameters
       ----------
       custom_function
         A custom user-supplied function that returns the
         functions that the user would like to be called for
         each step.
       setup: bool
         Whether or not to return the functions used to setup the
         space and input for the advance_XXX functions returned by
         this iterator. This is called once at the start of a run
         to return the functions that must be called to setup the
         model

       Returns
       -------
       funcs: List[function]
         The list of functions that ```iterate``` will call in sequence
    """

    kwargs["setup"] = setup

    if setup:
        # Return the functions needed to initialise this iterator
        core_funcs = iterate_core(**kwargs)

        if iterator_needs_setup(custom_function):
            custom_funcs = custom_function(**kwargs)
        else:
            custom_funcs = None

    else:
        core_funcs = iterate_core(**kwargs)
        custom_funcs = custom_function(**kwargs)

    # make sure that the core functions are called, and that
    # their call is before the custom function (unless the user
    # has moved them, which we hope was for a good reason!)
    if core_funcs is None or len(core_funcs) == 0:
        if custom_funcs is None:
            return []
        else:
            return custom_funcs

    elif custom_funcs is None or len(custom_funcs) == 0:
        return core_funcs

    else:
        for i in range(len(core_funcs)-1, -1, -1):
            # move backwards so that the first custom function
            # is prepended last
            if core_funcs[i] not in custom_funcs:
                custom_funcs.insert(0, core_funcs[i])

        return custom_funcs




from typing import List as _List
from typing import Dict as _Dict

__all__ = ["VariableSets", "VariableSet"]


class VariableSet:
    """This class holds a single set of adjustable variables that
       are used to adjust the variables as part of a model run

       Examples
       --------
       >>> v = VariableSet()
       >>> v["beta[1]"] = 0.95
       >>> v["beta[2]"] = 0.9
       >>> print(v.fingerprint())
       (beta[1]=0.95, beta[2]=0.9)[repeat 1]
       >>> params = Parameters()
       >>> params.set_disease("ncov")
       >>> v.adjust(params)
       >>> print(params.disease_params.beta[1],
       >>>       params.disease_params.beta[2])
       0.95 0.9
    """
    def __init__(self,
                 variables: _Dict[str, float] = None,
                 repeat_index: int = 1,
                 names: _List[str] = None,
                 values: _List[float] = None):
        """Construct a new VariableSet from the passed adjusted variable
           values.

           Parameters
           ----------
           names: List[str]
             The list of the names of the variables to adjust
           values: List[float]
             The values of the variables to adjust (same order as names)
           variables: Dict[str, float]
             names and values of variables to adjust passed as a dictionary
           repeat_index: int
             the index used to distinguish different repeats of the same
             VariableSet from one another

           Examples
           --------
           >>> v = VariableSet()
           >>> v["beta[1]"] = 0.95
           >>> v["beta[2]"] = 0.9
           >>> print(v.fingerprint())
           (beta[1]=0.95, beta[2]=0.9)[repeat 1]
        """
        self._names = None
        self._vals = None
        self._varnames = None
        self._varidxs = None
        self._idx = None

        if variables is not None:
            for name, value in variables.items():
                self._add(name, value)

        if values is not None:
            if names is None:
                names = ["beta[2]", "beta[3]", "progress[1]",
                         "progress[2]", "progress[3]"]

            if len(names) != len(values):
                raise IndexError(
                    f"The number of variable values '{values}' must equal "
                    f"the number of variable names '{names}'")

            for name, value in zip(names, values):
                self._add(name, value)

        self._idx = repeat_index

    def __str__(self):
        """Return a printable representation of the variables to
           be adjusted
        """
        s = []

        if self._vals is not None and len(self._vals) > 0:
            for key, value in zip(self._names, self._vals):
                s.append(f"{key}={value}")

        if len(s) == 0:
            return f"(NO_CHANGE)[repeat {self._idx}]"
        else:
            return f"({', '.join(s)})[repeat {self._idx}]"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, dict):
            other = VariableSet(variables=other)

        if self._idx != other._idx:
            return False

        if len(self) != len(other):
            return False

        if self._vals is None:
            return

        for i in range(0, len(self._vals)):
            if self._vals[i] != other._vals[i]:
                return False

            if self._varnames[i] != other._varnames[i]:
                return False

            if self._varidxs[i] != other._varidxs[i]:
                return False

            if self._names[i] != other._names[i]:
                return False

        return True

    def __len__(self):
        if self._vals is None:
            return 0
        else:
            return len(self._vals)

    def __getitem__(self, key):
        if self._vals is None:
            raise KeyError(f"No adjustable parameter {key} in an empty set")

        for i, name in enumerate(self._names):
            if key == name:
                return self._vals[i]

        raise KeyError(f"No adjustable parameter {key}. Available parameters "
                       f"are '{self._names}'")

    def __setitem__(self, key, value):
        if self._names is None:
            self._names = []

        for i, name in enumerate(self._names):
            if key == name:
                self._vals[i] = value
                return

        self._add(key, value)

    def _add(self, name, value):
        """Internal function to add a new variable called 'name' to
           be varied - it will be set equal to 'value'
        """
        import re

        if self._vals is None:
            self._names = []
            self._vals = []
            self._varnames = []
            self._varidxs = []

        name = name.strip()

        # look for 'variable[index]'
        m = re.search(r"(\w+)\[(\d+)\]", name)

        if m:
            self._varnames.append(m.group(1))
            self._varidxs.append(int(m.group(2)))
            self._names.append(name)
            self._vals.append(float(value))
        else:
            self._varnames.append(name)
            self._varidxs.append(None)
            self._names.append(name)
            self._vals.append(float(value))

    def variable_names(self):
        """Return the names of the variables that will be adjusted
           by this VariableSet

           Returns
           -------
           names: List[str]
             The list of names of variables to be adjusted
        """
        if self._vals is None or len(self._vals) == 0:
            return None
        else:
            from copy import deepcopy
            return deepcopy(self._names)

    def variable_values(self):
        """Return the values that the variables will be adjusted to.
           Note that 'None' means that the variable won't be adjusted
           from its default (original) value

           Returns
           -------
           values: List[float]
             The list of values for variables to be adjusted to
        """
        if self._vals is None or len(self._vals) == 0:
            return None
        else:
            from copy import deepcopy
            return deepcopy(self._vals)

    def variables(self):
        """Return the variables (name and values) to be adjusted

           Returns
           -------
           variables: Dict[str, float]
             The dictionary mapping the names of the variables that
             with be adjusted to their desired values
        """
        v = {}

        for name, value in zip(self._names, self._vals):
            v[name] = value

        return v

    def repeat_index(self):
        """Return the repeat index of this set. The repeat index is the
           ID of this set if the VariableSet is repeated. The index should
           range from 1 to nrepeats

           Returns
           -------
           index: int
             The repeat index of this set
        """
        return self._idx

    def make_compatible_with(self, other):
        """Return a copy of this VariableSet which has been made
           compatible with 'other'. This means that it will change
           the same variables as 'other', e.g. by adding 'None'
           changes for missing variables. This will raise an error
           if it is not possible to make this set compatible

           Parameters
           ----------
           other: VariableSet
             The passed VariableSet for which this should be made compatible

           Returns
           -------
           result: VariableSet
             A copy of this VariableSet which is now compatible with 'other'

           Example
           -------
           >>> v1 = VariableSet()
           >>> v1["beta1"] = 0.9
           >>> v1["beta2"] = 0.8

           >>> v2 = VariableSet()
           >>> v2["beta1"] = 0.6
           >>> v2 = v2.make_compatible_with(v1)
           >>> print(v2)
           (beta1=0.6, beta2=0.8)[repeat 1]
        """
        from copy import deepcopy

        if self._names is None:
            v = deepcopy(other)
            v._idx = self._idx
            return v

        if other._names is None:
            raise ValueError(f"VariableSet {self} is not compatible with "
                             f"VariableSet {other}")

        nmatch = 0

        for name in self._names:
            if name not in other._names:
                raise ValueError(f"VariableSet {self} is not compatible with "
                                 f"VariableSet {other}")
            nmatch += 1

        if len(other._names) == nmatch:
            # fully compatible
            return deepcopy(self)

        v = deepcopy(self)

        for name in other._names:
            if name not in self._names:
                v[name] = other[name]

        return v

    @staticmethod
    def extract_values(fingerprint: str):
        """Return the original values from the passed fingerprint
           or filename. This assumes that the fingerprint
           was created using the 'fingerprint' function, namely
           that any integers are actually 0.INTEGER

           Parameters
           ----------
           fingerprint: str
             The fingerprint (or filename) to decode

           Returns
           -------
           (values, repeat): (List[float], int)
             The list of values of the variables and the repeat
             index. The repeat index is None if it wasn't included
             in the fingerprint
        """
        from pathlib import Path

        # if this is a filename then extract only the fingerprint
        fingerprint = Path(Path(fingerprint).name)
        for suffix in fingerprint.suffixes:
            fingerprint = str(fingerprint).replace(suffix, "")

        # does this have a repeat index - if so, this is the last
        # value at the end after the hyphen (negative numbers are ~)
        parts = fingerprint.split("-")

        if len(parts) > 1:
            repeat_index = int(parts[-1])
            fingerprint = "-".join(parts[0:-1])
        else:
            repeat_index = None

        # now get the values
        values = []

        for part in fingerprint.split("_"):
            try:
                if part.startswith("~"):
                    scl = -1.0
                    part = part[1:]
                else:
                    scl = 1.0

                if part.find(".") != -1:
                    # this is already a floating point number
                    value = float(part)
                else:
                    # this is an integer, which represents 0.INT
                    part = f"0.{part}"
                    value = float(part)

                values.append(value)
            except Exception:
                # this is not part of the fingerprint
                pass

        return (values, repeat_index)

    def fingerprint(self, include_index: bool = False):
        """Return a fingerprint for this VariableSet. This can be
           used to quickly identify and distinguish the values of
           the variables in this set from the values in other
           VariableSets which have the same adjustable variables,
           but different parameters

           Parameters
           ----------
           include_index: bool
             Whether or not to include the repeat_index in the fingerprint

           Returns
           -------
           fingerprint: str
             The fingerprint for this VariableSet
        """
        if self._vals is None or len(self._vals) == 0:
            f = "NO_CHANGE"
        else:
            f = None
            for val in self._vals:
                v = str(val)
                if v.startswith("-"):
                    v = v[1:]
                if v.startswith("0"):
                    v = v[1:]
                if v.startswith("."):
                    v = v[1:]

                if val < 0:
                    # use `~` for negative numbers
                    v = "~" + v

                if f is None:
                    f = v
                else:
                    f += "_" + v

        if include_index:
            return "%s-%03d" % (f, self._idx)
        else:
            return f

    def adjust(self, params):  # should be 'Parameters' but circular include
        """Use the variables in this set to adjust the passed parameters.
           Note that this directly modifies 'params'

           Parameters
           ----------
           params: Parameters
             The parameters whose variables will be adjusted

           Returns
           -------
           None

           Examples
           --------
           >>> v = VariableSet()
           >>> v["beta[1]"] = 0.95
           >>> v["beta[2]"] = 0.9
           >>> print(v.fingerprint())
           (beta[1]=0.95, beta[2]=0.9)[repeat 1]
           >>> params = Parameters()
           >>> params.set_disease("ncov")
           >>> v.adjust(params)
           >>> print(params.disease_params.beta[1],
           >>>       params.disease_params.beta[2])
           0.95 0.9
        """
        if self._vals is None or len(self._vals) == 0:
            return

        try:
            for varname, varidx, value in zip(self._varnames, self._varidxs,
                                              self._vals):
                if value is not None:
                    if varname == "beta":
                        params.disease_params.beta[varidx] = value
                    elif varname == "progress":
                        params.disease_params.progress[varidx] = value
                    elif varname == "too_ill_to_move":
                        params.disease_params.too_ill_to_move[varidx] = value
                    elif varname == "contrib_foi":
                        params.disease_params.contrib_foi[varidx] = value
                    else:
                        raise KeyError(
                            f"Cannot set unrecognised parameter {varname} "
                            f"to {value}")
        except Exception as e:
            raise ValueError(
                f"Unable to set parameters from {self}. Error "
                f"equals {e.__class__}: {e}")


class VariableSets:
    """This class holds the collection of all VariableSet objects
       that contain the set of adjustable variables that are used
       to control a single run of the model

       Examples
       --------
       >>> v = VariableSets()
       >>> v.append({"beta[2]": 0.95, "beta[3]": 0.9})
       >>> v.append({"beta[1]": 0.86, "beta[2]": 0.89})
       >>> print(v)
       {(beta[2]=0.95, beta[3]=0.9)[repeat 1], (beta[1]=0.86,
       beta[2]=0.89)[repeat 1]}
       >>> v = v.repeat(2)
       >>> print(v)
       {(beta[2]=0.95, beta[3]=0.9)[repeat 1], (beta[1]=0.86,
       beta[2]=0.89)[repeat 1], (beta[2]=0.95, beta[3]=0.9)[repeat 2],
       (beta[1]=0.86, beta[2]=0.89)[repeat 2]}
    """
    def __init__(self):
        """Initialise an empty VariableSets object

           Parameters
           ----------
           None

           Returns
           -------
           None
        """
        self._vars = []

    def __str__(self):
        s = []

        for v in self._vars:
            s.append(str(v))

        if len(s) == 1:
            return s[0]
        elif len(s) > 0:
            return "{" + ", ".join(s) + "}"
        else:
            return "VariableSets:empty"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, dict):
            v = VariableSet(variables=other)
            other = VariableSets()
            other.append(v)

        if len(self._vars) != len(other._vars):
            return False

        for v0, v1 in zip(self._vars, other._vars):
            if v0 != v1:
                return False

        return True

    def __len__(self):
        if self._vars is None:
            return 0
        else:
            return len(self._vars)

    def __getitem__(self, i: int):
        """Return the VariableSet at the specified index"""
        if self._vars is None:
            raise IndexError("Cannot index an empty VariableSets object")
        else:
            return self._vars[i]

    def append(self, variables: VariableSet):
        """Append the passed set of variables to the set that will
           be used to run a model. If there are any existing
           VariableSet objects in this list, then the new VariableSet
           must adjust the same variables

           Parameters
           ----------
           variables: VariableSet
             The VariableSet to append to this list. If you pass a
             dict of {str: float} values, then this will automatically
             be converted into a VariableSet. Note that all VariableSet
             objects in a VariableSets must adjust the same variables

           Returns
           -------
           None
        """
        if isinstance(variables, dict):
            variables = VariableSet(variables=variables)

        if self._vars is None:
            self._vars = []

        if len(self._vars) > 0:
            variables = variables.make_compatible_with(self._vars[0])

        self._vars.append(variables)

    def repeat(self, nrepeats: int):
        """Return a copy of this VariableSet in which all of the
           unique VaribleSet objects have been repeated 'nrepeats'
           times

           Parameters
           ----------
           nrepeats: int
             The number of repeats of the VariableSet objects to
             perform

           Returns
           -------
           repeats: VariableSets
             A new VariableSets object containing 'nrepeats' copies
             of the VariableSet objects from this set
        """
        if nrepeats <= 1:
            return self

        from copy import deepcopy

        repeats = VariableSets()

        for i in range(1, nrepeats+1):
            for v in self._vars:
                v2 = deepcopy(v)
                v2._idx = i
                repeats.append(v2)

        return repeats

    @staticmethod
    def read(filename: str, line_numbers: _List[int] = None):
        """Read and return collection of VariableSet objects from the
           specified line number(s) of the specified file

           Parameters
           ----------
           filename: str
             The name of the file from which to read the VariableSets
           line_numbers: List[int]
             The line numbers from the file to read. This is 0-indexed,
             meaning that the first line is line 0. If this is None,
             then all lines are read and used

           Returns
           -------
           variables: VariableSets
             The collection of VariableSet objects that have been read,
             in the order they were read from the file
        """
        if not isinstance(line_numbers, list):
            if line_numbers is not None:
                line_numbers = [line_numbers]

        variables = VariableSets()

        i = -1
        with open(filename, "r") as FILE:
            line = FILE.readline()

            # find the first line of the file. Use this to work out
            # the separator to use and also to see if the user has
            # named the adjustable variables
            first_line = None
            separator = ","

            # default adjustable variables
            titles = ["beta[2]", "beta[3]", "progress[1]",
                      "progress[2]", "progress[3]"]

            while line:
                i += 1

                line = line.strip()

                if first_line is None:
                    if len(line) > 0:
                        # this is a valid first line - what separator
                        # should we use?
                        if line.find(",") != -1:
                            separator = ","
                        else:
                            separator = None  # spaces

                        first_line = line

                        words = line.split(separator)

                        try:
                            float(words[0])
                            is_title_line = False
                        except Exception:
                            is_title_line = True

                        if is_title_line:
                            titles = words
                            line = FILE.readline()
                            continue

                if line_numbers is None or i in line_numbers:
                    words = line.split(separator)

                    if len(words) != len(titles):
                        raise ValueError(
                            f"Corrupted input file. Expecting {len(titles)} "
                            f"values. Received {line}")

                    vals = []

                    try:
                        for word in words:
                            vals.append(float(word))
                    except Exception:
                        raise ValueError(
                               f"Corrupted input file. Expected {len(titles)} "
                               f"numbers. Received {line}")

                    variables.append(VariableSet(names=titles,
                                                 values=vals))

                    if line_numbers is not None:
                        if len(variables) == len(line_numbers):
                            return variables

                line = FILE.readline()

        # get here if we can't find this line in the file (or if we
        # are supposed to read all lines)
        if line_numbers is None:
            return variables
        else:
            raise ValueError(
                    f"Cannot read parameters from line {line_numbers} "
                    f"as the number of lines in the file is {i+1}")




from dataclasses import dataclass as _dataclass
from typing import List as _List
from copy import deepcopy as _deepcopy
import pathlib as _pathlib
import os as _os
import json as _json

from ._inputfiles import InputFiles
from ._disease import Disease
from ._variableset import VariableSets, VariableSet


__all__ = ["Parameters", "get_repository_version"]

_default_parameters_path = _os.path.join(_pathlib.Path.home(),
                                         "GitHub", "MetaWardsData")

_default_folder_name = "parameters"


_repositories = {}


def generate_repository_version(repository):
    """Try to run the './version' script within the passed repository,
       to generate the required 'version.txt' file
    """
    import subprocess
    script = _os.path.join(repository, "version")
    print(f"Regenerating version information using {script}")
    subprocess.run(script, cwd=repository)


def get_repository_version(repository: str):
    """Read and return the Git version of the passed repository

       Parameters
       ----------
       repository: str
         The full path to the repository whose version should be obtained

       Returns
       -------
       version_data: dict
         A dictionary containing version information for the repository
    """
    global _repositories

    if repository in _repositories:
        return _repositories[repository]

    filename = _os.path.join(repository, "version.txt")

    try:
        with open(filename) as FILE:
            version = _json.load(FILE)
            _repositories[repository] = version
            return version
    except Exception:
        pass

    # could not get the version, so see if we have permission
    # to run the 'version' program
    try:
        generate_repository_version(repository)

        with open(filename) as FILE:
            version = _json.load(FILE)
            _repositories[repository] = version
            return version
    except Exception:
        print(f"Could not find the repository version info in {filename}."
              f"Please make sure that you have run './version' in that "
              f"repository to generate the version info.")
        _repositories[repository] = {"repository": "unknown",
                                     "version": "unknown",
                                     "branch": "unknown"}
        return _repositories[repository]


@_dataclass
class Parameters:
    """The full set of Parameters that are used to control the model
       outbreak over a Network. The combination of a Network and
       a Parameters defines the model outbreak.

       Load the Parameters using the Parameters.load function, and
       then add extra data using the various "set" and "add" functions,
       e.g.

       Examples
       --------
       >>> params = Parameters.load("march29")
       >>> params.set_disease("ncov")
       >>> params.set_input_files("2011Data")
       >>> params.add_seeds("ExtraSeedsBrighton.dat")
    """
    #: The set of input files that define the model Network
    input_files: InputFiles = None
    #: The name of the UV file
    uv_filename: str = None
    #: The set of parameters that define the disease
    disease_params: Disease = None

    #: The set of files that contain additional seeds that
    #: seed the outbreak during the model run
    additional_seeds: _List[str] = None

    #: The fraction of day considered "day" for work, e.g. 0.7 * 24 hours
    length_day: float = 0.7
    #: The fraction of day considered "day" for play
    plength_day: float = 0.5
    #: The number of initial infections
    initial_inf: int = 5

    static_play_at_home: float = 0.0
    dyn_play_at_home: float = 0.0

    data_dist_cutoff: float = 10000000.0
    dyn_dist_cutoff: float = 10000000.0

    play_to_work: float = 0.0
    work_to_play: float = 0.0

    local_vaccination_thresh: int = 4
    global_detection_thresh: int = 4
    daily_ward_vaccination_capacity: int = 5
    neighbour_weight_threshold: float = 0.0

    #: proportion of daily imports
    daily_imports: float = 0.0
    UV: float = 0.0

    _name: str = None
    _version: str = None
    _authors: str = None
    _contacts: str = None
    _references: str = None
    _filename: str = None
    _repository: str = None
    _repository_version: str = None
    _repository_branch: str = None
    _repository_dir: str = None

    def __str__(self):
        return f"Parameters {self._name}\n" \
               f"loaded from {self._filename}\n" \
               f"version: {self._version}\n" \
               f"author(s): {self._authors}\n" \
               f"contact(s): {self._contacts}\n" \
               f"references(s): {self._references}\n" \
               f"repository: {self._repository}\n" \
               f"repository_branch: {self._repository_branch}\n" \
               f"repository_version: {self._repository_version}\n\n" \
               f"length_day = {self.length_day}\n" \
               f"plength_day = {self.plength_day}\n" \
               f"initial_inf = {self.initial_inf}\n" \
               f"static_play_at_home = {self.static_play_at_home}\n" \
               f"dyn_play_at_home = {self.dyn_play_at_home}\n" \
               f"data_dist_cutoff = {self.data_dist_cutoff}\n" \
               f"dyn_dist_cutoff = {self.dyn_dist_cutoff}\n" \
               f"play_to_work = {self.play_to_work}\n" \
               f"work_to_play = {self.work_to_play}\n" \
               f"local_vaccination_thresh = " \
               f"{self.local_vaccination_thresh}\n" \
               f"global_detection_thresh = {self.global_detection_thresh}\n" \
               f"daily_ward_vaccination_capacity = " \
               f"{self.daily_ward_vaccination_capacity}\n" \
               f"neighbour_weight_threshold = " \
               f"{self.neighbour_weight_threshold}\n" \
               f"daily_imports = {self.daily_imports}\n" \
               f"UV = {self.UV}\n" \
               f"additional_seeds = {self.additional_seeds}\n\n"

    @staticmethod
    def get_repository(repository: str = None):
        """Return the repository location and version information
           for the passed repository

           Parameters
           ----------
           repository: str
             Location on the filesystem of the repository. If this
             is None then it will be searched for using first
             the environment variable METAWARDSDATA, then
             $HOME/GitHub/MetaWardsData, then ./METAWARDSDATA

           Returns
           -------
           (repository, version): tuple
             A tuple of the location on disk of the repository,
             plus the version information (git ID etc)
        """
        if repository is None:
            repository = _os.getenv("METAWARDSDATA")
            if repository is None:
                repository = _default_parameters_path

        from pathlib import Path
        import os
        repository = os.path.expanduser(os.path.expandvars(repository))
        repository = Path(repository).absolute().resolve()

        if not os.path.exists(repository):
            raise FileNotFoundError(
                    f"Cannot find the MetaWardsData repository "
                    f"at {repository}")

        if not os.path.isdir(repository):
            raise FileNotFoundError(
                    f"Expected {repository} to be a directory containing "
                    f"the MetaWardsData repository. It isn't?")

        v = get_repository_version(repository)
        return (repository, v)

    @staticmethod
    def load(parameters: str = "march29",
             repository: str = None,
             folder: str = _default_folder_name,
             filename: str = None):
        """This will return a Parameters object containing all of the
           parameters loaded from the parameters found in file
           f"{repository}/{folder}/{parameters}.json"

           By default this will load the march29 parameters from
           $HOME/GitHub/model_data/2011Data/parameters/march29.json

           Alternatively, you can provide the exact path to the
           filename via the 'filename' argument

           Parameters
           ----------
           parameters: str
             The name of the parameters to load. This is the name that
             will be searched for in the METAWARDSDATA parameters directory
           repository: str
             The location of the cloned METAWARDSDATA repository
           folder: str
             The name of the folder within the METAWARDSDATA repository
             that contains the parameters
           filename: str
             The name of the file to load the parameters from - this directly
             loads this file without searching through the METAWARDSDATA
             repository

           Returns
           -------
           params: Parameters
             The constructed and validated parameters
        """
        repository_version = None
        repository_branch = None
        repository_dir = None

        if filename is None:
            (repository, v) = Parameters.get_repository(repository)
            filename = _os.path.join(repository, folder, f"{parameters}.json")
            repository_dir = repository
            repository = v["repository"]
            repository_branch = v["branch"]
            repository_version = v["version"]

        json_file = filename

        try:
            with open(json_file, "r") as FILE:
                import json
                data = json.load(FILE)

        except Exception as e:
            print(f"Could not find the parameters file {json_file}")
            print(f"Either it does not exist of was corrupted.")
            print(f"Error was {e.__class__} {e}")
            print(f"To download the parameters type the command:")
            print(f"  git clone https://github.com/metawards/MetaWardsData")
            print(f"and then re-run this function passing in the full")
            print(f"path to where you downloaded this directory")
            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")

        par = Parameters(
                length_day=data["length_day"],
                plength_day=data["plength_day"],
                initial_inf=data["initial_inf"],
                static_play_at_home=data["static_play_at_home"],
                dyn_play_at_home=data["dyn_play_at_home"],
                data_dist_cutoff=data["data_dist_cutoff"],
                dyn_dist_cutoff=data["dyn_dist_cutoff"],
                play_to_work=data["play_to_work"],
                work_to_play=data["work_to_play"],
                local_vaccination_thresh=data["local_vaccination_threshold"],
                global_detection_thresh=data["global_detection_threshold"],
                daily_ward_vaccination_capacity=data[
                                "daily_ward_vaccination_capacity"],
                neighbour_weight_threshold=data["neighbour_weight_threshold"],
                daily_imports=data["daily_imports"],
                UV=data["UV"],
                _name=data["name"],
                _authors=data["author(s)"],
                _version=data["version"],
                _contacts=data["contact(s)"],
                _references=data["reference(s)"],
                _filename=json_file,
                _repository=repository,
                _repository_dir=repository_dir,
                _repository_branch=repository_branch,
                _repository_version=repository_version
                )

        print("Using parameters:")
        print(par)

        return par

    def add_seeds(self, filename: str):
        """Add an 'additional seeds' file that can be used to
           seed wards with new infections at different times and
           locations. Several additional_seed files can be added

           Parameters
           ----------
           filename: str
             Name of the file containing the additional seeds
        """
        # resolve the filename to the GitHub repo if possible...
        if self.additional_seeds is None:
            self.additional_seeds = []

        if not _os.path.exists(filename):
            f = _os.path.join(self._repository_dir, "extra_seeds", filename)

            if _os.path.exists(f):
                filename = f
            else:
                raise FileExistsError(
                        f"Unable to find extra seeds file {filename} in "
                        f"the current directory or in {f}")

        self.additional_seeds.append(filename)

    def set_input_files(self, input_files: InputFiles):
        """Set the input files that are used to initialise the
           simulation

           Parameters
           ----------
           input_files: InputFiles
             The set of input files that will be used to load the Network.
             If a string is passed then the InputFiles will be loaded
             based on that string.
        """
        if isinstance(input_files, str):
            input_files = InputFiles.load(input_files,
                                          repository=self._repository_dir)

        print("Using input files:")
        print(input_files)

        self.input_files = _deepcopy(input_files)

    def set_disease(self, disease: Disease):
        """"Set the disease that will be modelled

            Parameters:
              disease: The disease to be modelled. If a string is passed
              then the disease will be loaded using that string
        """
        if isinstance(disease, str):
            disease = Disease.load(disease,
                                   repository=self._repository_dir)

        print("Using disease")
        print(disease)

        self.disease_params = _deepcopy(disease)

    def set_variables(self, variables: VariableSet):
        """This function sets the adjustable variable values to those
           specified in 'variables' in A COPY OF THIS PARAMETERS OBJECT.
           This returns the copy. It does not change this object

           Parameters
           ----------
           variables: VariableSet
             The variables that will be adjusted before the model run.
             This adjusts the parameters and returns them in a deep copy

           Returns
           -------
           params: Parameters
             A copy of this set of parameters with the variables adjusted
        """
        params = _deepcopy(self)

        if isinstance(variables, dict):
            variables = VariableSet(variables)

        variables.adjust(params)

        return params

    @staticmethod
    def read_variables(filename: str, line_numbers: _List[int]):
        """Read in extra variable parameters from the specified line number(s)
           of the specified file, returning the list
           of the dictionaries of variables that have been
           read. You can then apply those variable parameters
           using the 'set_variables' function

           Parameters
           ----------
           filename: str
             The file from which to read the adjustable variables
           line_numbers: List[int]
             All of the line numbers from which to read. If this is
             None then all lines will be read.

           Returns
           -------
           variables: VariableSets
             The VariableSets containing all of the adjustable variables
        """
        return VariableSets.read(filename, line_numbers)




from .._network import Network
from .._parameters import Parameters
from ._profiler import Profiler

__all__ = ["iterate_weekend"]


def iterate_weekend(network: Network, infections, play_infections,
                    params: Parameters, rngs, timestep: int,
                    population: int, nthreads: int = None,
                    profiler: Profiler = None,
                    is_dangerous=None, SELFISOLATE: bool = False,
                    ):
    """Iterate the model forward one timestep (day) using the supplied
       network and parameters, advancing the supplied infections,
       and using the supplied random number generator (rngs)
       (array, with one generator per thread)
       to generate random numbers. This iterates for a non-working
       (weekend) day (with only random movements)

       If SELFISOLATE is True then you need to pass in
       is_dangerous, which should be an array("i", network.nnodes)
    """
    raise AssertionError("Will write iterate_weekend later...")




from dataclasses import dataclass

__all__ = ["Link"]


@dataclass
class Link:
    """This is a link between nodes in the network"""
    ifrom: int = None        # integers (indexes) of origins and destinations
    ito: int = None

    weight: float = None     # weight of link, used to save the original
                             # number of susceptibles in work matrix

    suscept: float = None    # number of susceptibles in the case of the
                             # work matrix. Used to save the weight of
                             # the play matrix

    distance: float = None   # the distance between two wards
    A: int = None            # Age index...




from .._network import Network
from ._profiler import Profiler, NullProfiler
from .._population import Population
from .._outputfiles import OutputFiles
from ._workspace import Workspace

__all__ = ["extract_data"]


def extract_data(network: Network, population: Population,
                 workspace: Workspace, output_dir: OutputFiles,
                 infections, play_infections, rngs, get_output_functions,
                 nthreads: int, profiler: Profiler = None):
    """Extract data from the network and write this to the specified
       output directory. Like :meth:`~metawards.utils.iterate` this
       uses a dynamic set of functions that can be utilised to
       customise what is output dynamically throughout the model
       run.

       Parameters
       ----------
       network: Network
         The network over which the model is being run
       population: Population
         The population experiencing the model outbreak
       workspace: Workspace
         A scratch-space that can be used to accumulate data while
         it is being extracted
       output_dir: OutputFiles
         The output directory to which to write all files
       infections
         Space to hold all of the 'work' infections
       play_infections
         Space to hold all of the 'play' infections
       rngs
         Thread-safe random number generators (one per thread)
       get_output_functions
         A function that should return the list of output functions
         that are called in sequence to write the output data.
         See :meth:`~metawards.extractors.extract_default` for an
         example of a suitable function
       nthreads: int
         The number of threads over which to parallelise extracting
         the output
       profiler: Profiler
         The profiler used to profile extracting the output
    """
    if profiler is None:
        profiler = NullProfiler()

    p = profiler.start("extract_data")

    output_functions = get_output_functions(population=population,
                                            nthreads=nthreads)

    for output_function in output_functions:
        output_function(network=network,
                        population=population,
                        workspace=workspace,
                        output_dir=output_dir,
                        infections=infections,
                        play_infections=play_infections,
                        rngs=rngs, nthreads=nthreads,
                        profiler=p)

    p.stop()



# Package setup copied from
# https://github.com/FedericoStra/cython-package-example
# Thanks - this was really helpful :-)

import os
import versioneer
from setuptools import setup, Extension

try:
    from Cython.Build import cythonize
    have_cython = True
except Exception:
    have_cython = False

cflags = '-O3 -march=native -Wall -fopenmp'

nbuilders = int(os.getenv("CYTHON_NBUILDERS", 2))

compiler_directives = {"language_level": 3, "embedsignature": True,
                       "boundscheck": False, "cdivision": True,
                       "initializedcheck": False, "cdivision_warnings": False,
                       "wraparound": False, "binding": False,
                       "nonecheck": False, "overflowcheck": False}

if os.getenv("CYTHON_LINETRACE", 0):
    print("Compiling with Cython line-tracing support - will be SLOW")
    define_macros = [("CYTHON_TRACE", "1")]
    compiler_directives["linetrace"] = True
else:
    define_macros = []

random_sources = ["src/metawards/ran_binomial/mt19937.c",
                  "src/metawards/ran_binomial/distributions.c"]


# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

# Currently each file compiles to its own module. Ideally I'd like
# them all to compile into a single module library, but this doesn't
# work (get the error "ImportError: dynamic module does not define
# init function" and the answers on stackoverflow didn't work.
# If anyone can help please do :-)


extensions = [
    Extension("metawards._nodes", ["src/metawards/_nodes.pyx"],
              define_macros=define_macros),
    Extension("metawards._links", ["src/metawards/_links.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._workspace",
              ["src/metawards/utils/_workspace.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._clear_all_infections",
              ["src/metawards/utils/_clear_all_infections.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._build_wards_network",
              ["src/metawards/utils/_build_wards_network.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._add_wards_network_distance",
              ["src/metawards/utils/_add_wards_network_distance.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._get_min_max_distances",
              ["src/metawards/utils/_get_min_max_distances.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._reset_everything",
              ["src/metawards/utils/_reset_everything.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._rescale_matrix",
              ["src/metawards/utils/_rescale_matrix.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._recalculate_denominators",
              ["src/metawards/utils/_recalculate_denominators.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._move_population",
              ["src/metawards/utils/_move_population.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._fill_in_gaps",
              ["src/metawards/utils/_fill_in_gaps.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._build_play_matrix",
              ["src/metawards/utils/_build_play_matrix.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._array", ["src/metawards/utils/_array.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._ran_binomial",
              ["src/metawards/utils/_ran_binomial.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.utils._assert_sane_network",
              ["src/metawards/utils/_assert_sane_network.pyx"],
              define_macros=define_macros),
    Extension("metawards.utils._get_array_ptr",
              ["src/metawards/utils/_get_array_ptr.pyx"],
              define_macros=define_macros),
    Extension("metawards.iterators._advance_play",
              ["src/metawards/iterators/_advance_play.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.iterators._advance_fixed",
              ["src/metawards/iterators/_advance_fixed.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.iterators._advance_infprob",
              ["src/metawards/iterators/_advance_infprob.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.iterators._advance_recovery",
              ["src/metawards/iterators/_advance_recovery.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.iterators._advance_foi",
              ["src/metawards/iterators/_advance_foi.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.iterators._advance_imports",
              ["src/metawards/iterators/_advance_imports.pyx"]+random_sources,
              define_macros=define_macros),
    Extension("metawards.extractors._output_default",
              ["src/metawards/extractors/_output_default.pyx"],
              define_macros=define_macros),
]

CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 0)))

if not have_cython:
    CYTHONIZE = False

os.environ['CFLAGS'] = cflags

if CYTHONIZE:
    extensions = cythonize(extensions, compiler_directives=compiler_directives,
                           nthreads=nbuilders)
else:
    extensions = no_cythonize(extensions)

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

with open("requirements-dev.txt") as fp:
    dev_requires = fp.read().strip().split("\n")

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=extensions,
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
        "docs": ["sphinx", "sphinx-rtd-theme"]
    },
    entry_points={
        "console_scripts": [
           "metawards = metawards.app.run:cli",
           "metawards-plot = metawards.app.plot:cli"
        ]
    }
)




from dataclasses import dataclass as _dataclass
from typing import List as _List

from ._parameters import Parameters
from ._nodes import Nodes
from ._links import Links
from ._population import Population
from ._outputfiles import OutputFiles

__all__ = ["Network"]


@_dataclass
class Network:
    """This class represents a network of wards. The network comprises
       nodes (representing wards), connected with links which represent
       work (predictable) links. There are also additional links for
       play (unpredictable/random) and weekend
    """

    #: The list of nodes (wards) in the network
    nodes: Nodes = None
    #: The links between nodes (work)
    to_links: Links = None
    #: The links between nodes (play)
    play: Links = None
    #: The links between nodes (weekend)
    weekend: Links = None

    #: The number of nodes in the network
    nnodes: int = 0
    #: The number of links in the network
    nlinks: int = 0
    #: The number of play links in the network
    plinks: int = 0

    #: The maximum allowable number of nodes in the network
    max_nodes: int = 10050
    #: The maximum allowable number of links in the network
    max_links: int = 2414000

    #: To seed provides additional seeding information
    to_seed: _List[int] = None

    #: The parameters used to generate this network
    params: Parameters = None

    @staticmethod
    def build(params: Parameters,
              calculate_distances: bool = True,
              build_function=None,
              distance_function=None,
              max_nodes: int = 10050,
              max_links: int = 2414000,
              nthreads: int = 1,
              profile: bool = True,
              profiler=None):
        """Builds and returns a new Network that is described by the
           passed parameters. If 'calculate_distances' is True, then
           this will also read in the ward positions and add
           the distances between the links.

           Optionally you can supply your own function to build the network,
           by supplying 'build_function'. By default, this is
           metawards.utils.build_wards_network.

           Optionally you can supply your own function to read and
           calculate the distances by supplying 'build_function'.
           By default this is metawards.add_wards_network_distance

           The network is built in allocated memory, so you need to specify
           the maximum possible number of nodes and links. The memory buffers
           will be shrunk back after building.
        """
        if profile:
            if profiler is None:
                from .utils import Profiler
                p = Profiler()
            else:
                p = profiler
        else:
            from .utils import NullProfiler
            p = NullProfiler()

        p = p.start("Network.build")

        if build_function is None:
            from .utils import build_wards_network
            build_function = build_wards_network

        p = p.start("build_function")
        network = build_function(params=params,
                                 profiler=p,
                                 max_nodes=max_nodes,
                                 max_links=max_links,
                                 nthreads=nthreads)
        p = p.stop()

        # sanity-check that the network makes sense - there are specific
        # requirements for the data layout
        network.assert_sane()

        if calculate_distances:
            p = p.start("add_distances")
            network.add_distances(distance_function=distance_function,
                                  nthreads=nthreads, profiler=p)
            p = p.stop()

        if params.input_files.seed:
            from .utils import read_done_file
            p = p.start("read_done_file")
            to_seed = read_done_file(params.input_files.seed)
            nseeds = len(to_seed)

            print(to_seed)
            print(f"Number of seeds equals {nseeds}")
            network.to_seed = to_seed
            p = p.stop()

        # By default, we initialise the network ready for a run,
        # namely make sure everything is reset and the population
        # is at work
        print("Reset everything...")
        p = p.start("reset_everything")
        network.reset_everything(nthreads=nthreads, profiler=p)
        p = p.stop()

        print("Rescale play matrix...")
        p = p.start("rescale_play_matrix")
        network.rescale_play_matrix(nthreads=nthreads, profiler=p)
        p = p.stop()

        print("Move population from play to work...")
        p = p.start("move_from_play_to_work")
        network.move_from_play_to_work(nthreads=nthreads, profiler=p)
        p = p.stop()

        if not p.is_null():
            p = p.stop()
            print(p)

        return network

    def assert_sane(self):
        """Assert that this network is sane. This checks that the network
           is laid out correctly in memory and that it doesn't have
           anything unexpected. Checking here will prevent us from having
           to check every time the network is accessed
        """
        from .utils import assert_sane_network
        assert_sane_network(self)

    def add_distances(self, distance_function=None, nthreads: int = 1,
                      profiler=None):
        """Read in the positions of all of the nodes (wards) and calculate
           the distances of the links.

           Optionally you can specify the function to use to
           read the positions and calculate the distances.
           By default this is mw.utils.add_wards_network_distance
        """
        if distance_function is None:
            from .utils import add_wards_network_distance
            distance_function = add_wards_network_distance

        distance_function(self, nthreads=nthreads)

        # now need to update the dynamic distance cutoff based on the
        # maximum distance between nodes
        print("Get min/max distances...")
        (_mindist, maxdist) = self.get_min_max_distances(nthreads=nthreads)

        self.params.dyn_dist_cutoff = maxdist + 1

    def initialise_infections(self, nthreads: int = 1):
        """Initialise and return the space that will be used
           to track infections
        """
        from .utils import initialise_infections
        return initialise_infections(self)

    def initialise_play_infections(self, nthreads: int = 1):
        """Initialise and return the space that will be used
           to track play infections
        """
        from .utils import initialise_play_infections
        return initialise_play_infections(self)

    def get_min_max_distances(self, nthreads: int = 1,
                              profiler=None):
        """Calculate and return the minimum and maximum distances
           between nodes in the network
        """
        try:
            return self._min_max_distances
        except Exception:
            pass

        from .utils import get_min_max_distances
        self._min_max_distances = get_min_max_distances(network=self,
                                                        nthreads=nthreads)

        return self._min_max_distances

    def reset_everything(self, nthreads: int = 1,
                         profiler=None):
        """Resets the network ready for a new run of the model"""
        from .utils import reset_everything
        reset_everything(network=self, nthreads=nthreads, profiler=profiler)

    def update(self, params: Parameters,
               nthreads: int = 1, profile: bool = False):
        """Update this network with a new set of parameters.
           This is used to update the parameters for the network
           for a new run. The network will be reset
           and ready for a new run.
        """
        if profile:
            from .utils import Profiler
            p = Profiler()
        else:
            from .utils import NullProfiler
            p = NullProfiler()

        p = p.start("Network.update")

        self.params = params

        p = p.start("reset_everything")
        self.reset_everything(nthreads=nthreads, profiler=p)
        p = p.stop()

        p = p.start("rescale_play_matrix")
        self.rescale_play_matrix(nthreads=nthreads, profiler=p)
        p = p.stop()

        p = p.start("move_from_play_to_work")
        self.move_from_play_to_work(nthreads=nthreads, profiler=p)
        p = p.stop()

        p = p.stop()

        if profile:
            print(p)

    def rescale_play_matrix(self, nthreads: int = 1,
                            profiler=None):
        """Rescale the play matrix"""
        from .utils import rescale_play_matrix
        rescale_play_matrix(network=self, nthreads=nthreads, profiler=profiler)

    def move_from_play_to_work(self, nthreads: int = 1,
                               profiler=None):
        """Move the population from play to work"""
        from .utils import move_population_from_play_to_work
        move_population_from_play_to_work(network=self, nthreads=nthreads,
                                          profiler=profiler)

    def run(self, population: Population,
            output_dir: OutputFiles,
            seed: int = None,
            nsteps: int = None,
            profile: bool = True,
            s: int = None,
            nthreads: int = None,
            iterator=None,
            extractor=None,
            profiler=None):
        """Run the model simulation for the passed population.
           The random number seed is given in 'seed'. If this
           is None, then a random seed is used.

           All output files are written to 'output_dir'

           The simulation will continue until the infection has
           died out or until 'nsteps' has passed (keep as 'None'
           to prevent exiting early).

           s is used to select the 'to_seed' entry to seed
           the nodes

           Parameters
           ----------
           population: Population
             The initial population at the start of the model outbreak.
             This is also used to set start date and day of the model
             outbreak
           output_dir: OutputFiles
             The directory to write all of the output into
           seed: int
             The random number seed used for this model run. If this is
             None then a very random random number seed will be used
           nsteps: int
             The maximum number of steps to run in the outbreak. If None
             then run until the outbreak has finished
           profile: bool
             Whether or not to profile the model run and print out the
             results
           profiler: Profiler
             The profiler to use - a new one is created if one isn't passed
           s: int
             Index of the seeding parameter to use
           nthreads: int
             Number of threads over which to parallelise this model run
           iterator: function
             Function that is called at each iteration to get the functions
             that are used to advance the model
           extractor: function
             Function that is called at each iteration to get the functions
             that are used to extract data for analysis or writing to files
        """
        # Create the random number generator
        from .utils._ran_binomial import seed_ran_binomial, ran_binomial

        rng = seed_ran_binomial(seed=seed)

        # Print the first five random numbers so that we can
        # compare to other codes/runs, and be sure that we are
        # generating the same random sequence
        randnums = []
        for i in range(0, 5):
            randnums.append(str(ran_binomial(rng, 0.5, 100)))

        print(f"First five random numbers equal {', '.join(randnums)}")
        randnums = None

        if nthreads is None:
            from .utils._parallel import get_available_num_threads
            nthreads = get_available_num_threads()

        print(f"Number of threads used equals {nthreads}")

        from .utils._parallel import create_thread_generators
        rngs = create_thread_generators(rng, nthreads)

        # Create space to hold the results of the simulation
        print("Initialise infections...")
        infections = self.initialise_infections()

        print("Initialise play infections...")
        play_infections = self.initialise_play_infections()

        from .utils import run_model
        population = run_model(network=self,
                               population=population,
                               infections=infections,
                               play_infections=play_infections,
                               rngs=rngs, s=s, output_dir=output_dir,
                               nsteps=nsteps,
                               profile=profile, nthreads=nthreads,
                               profiler=profiler,
                               iterator=iterator, extractor=extractor)

        return population




from .._network import Network
from .._population import Population
from .._variableset import VariableSets
from .._outputfiles import OutputFiles

from contextlib import contextmanager as _contextmanager

import os as _os
import sys as _sys

__all__ = ["get_number_of_processes", "run_models", "redirect_output"]


def get_number_of_processes(parallel_scheme: str, nprocs: int = None):
    """This function works out how many processes have been set
       by the paralellisation system called 'parallel_scheme'
    """
    if nprocs is None:
        if parallel_scheme == "multiprocessing":
            return 1
        elif parallel_scheme == "mpi4py":
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            nprocs = comm.Get_size()
            return nprocs
        elif parallel_scheme == "scoop":
            raise ValueError(
                    f"You must specify the number of processes for "
                    f"scoop to parallelise over")
        else:
            raise ValueError(
                    f"You must specify the number of processes to "
                    f"use for parallel scheme '{parallel_scheme}'")

    if parallel_scheme == "mpi4py":
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        n = comm.Get_size()

        if n < nprocs:
            return n
        else:
            return nprocs
    elif parallel_scheme == "scoop":
        return 4
    elif parallel_scheme == "multiprocessing":
        return nprocs
    else:
        raise ValueError(
            f"Unrecognised parallelisation scheme {parallel_scheme}")


@_contextmanager
def redirect_output(outdir):
    """Nice way to redirect stdout and stderr - thanks to
       Emil Stenström in
       https://stackoverflow.com/questions/6735917/redirecting-stdout-to-nothing-in-python
    """
    new_out = open(_os.path.join(outdir, "output.txt"), "w")
    old_out = _sys.stdout
    _sys.stdout = new_out

    new_err = open(_os.path.join(outdir, "output.err"), "w")
    old_err = _sys.stderr
    _sys.stderr = new_err

    try:
        yield new_out
    finally:
        _sys.stdout = old_out
        _sys.stderr = old_err
        new_out.close()
        new_err.close()


def run_models(network: Network, variables: VariableSets,
               population: Population,
               nprocs: int, nthreads: int, seed: int,
               nsteps: int, output_dir: OutputFiles,
               iterator: str, extractor: str,
               profile: bool, parallel_scheme: str):
    """Run all of the models on the passed Network that are described
       by the passed VariableSets

       Parameters
       ----------
       network: Network
         The network to model
       variables: VariableSets
         The sets of VariableSet that represent all of the model
         runs to perform
       population: Population
         The initial population for all of the model runs. This also
         contains the starting date and day for the model outbreak
       nprocs: int
         The number of model runs to perform in parallel
       nthreads: int
         The number of threads to parallelise each model run over
       seed: int
         Random number seed which is used to generate random seeds
         for all model runs
       nsteps: int
         The maximum number of steps to perform for each model - this
         will run until the outbreak is over if this is None
       output_dir: OutputFiles
         The OutputFiles that represents the directory in which all
         output should be placed
       iterator: str
         Iterator to load that will be used to iterate the outbreak
       extractor: str
         Extractor to load that will be used to extract information
       profile: bool
         Whether or not to profile the model run and print out live
         timing (useful for performance debugging)
       parallel_scheme: str
         Which parallel scheme (multiprocessing, mpi4py or scoop) to use
         to run multiple model runs in parallel
    """

    # this variable is used to pick out some of the additional seeds?
    s = -1

    if len(variables) == 1:
        # no need to do anything complex - just a single run
        params = network.params.set_variables(variables[0])

        network.update(params, profile=profile)

        trajectory = network.run(population=population, seed=seed,
                                 s=s, nsteps=nsteps,
                                 output_dir=output_dir,
                                 iterator=iterator,
                                 extractor=extractor,
                                 profile=profile,
                                 nthreads=nthreads)

        return [(variables[0], trajectory)]

    # generate the random number seeds for all of the jobs
    # (for testing, we will use the same seed so that I can check
    #  that they are all working)
    seeds = []

    if seed == 0:
        # this is a special mode that a developer can use to force
        # all jobs to use the same random number seed (15324) that
        # is used for comparing outputs. This should NEVER be used
        # for production code
        print("** WARNING: Using special mode to fix all random number")
        print("** WARNING: seeds to 15324. DO NOT USE IN PRODUCTION!!!")

        for i in range(0, len(variables)):
            seeds.append(15324)

    else:
        from ._ran_binomial import seed_ran_binomial, ran_int
        rng = seed_ran_binomial(seed)

        # seed the rngs used for the sub-processes using this rng
        for i in range(0, len(variables)):
            seeds.append(ran_int(rng, 10000, 99999999))

    # set the output directories for all of the jobs - this is based
    # on the fingerprint, so should be unique for each job
    outdirs = []

    for v in variables:
        f = v.fingerprint(include_index=True)
        d = _os.path.join(output_dir.get_path(), f)
        outdirs.append(d)

    outputs = []

    print(f"\nRunning {len(variables)} jobs using {nprocs} process(es)")

    if nprocs == 1:
        # no need to use a pool, as we will repeat this calculation
        # several times
        for i, variable in enumerate(variables):
            seed = seeds[i]
            outdir = outdirs[i]

            with output_dir.open_subdir(outdir) as subdir:
                print(f"\nRunning parameter set {i+1} of {len(variables)} "
                      f"using seed {seed}")
                print(f"All output written to {subdir.get_path()}")

                with redirect_output(subdir.get_path()):
                    print(f"Running variable set {i+1}")
                    print(f"Variables: {variable}")
                    print(f"Random seed: {seed}")
                    print(f"nthreads: {nthreads}")

                    # no need to do anything complex - just a single run
                    params = network.params.set_variables(variable)

                    network.update(params, profile=profile)

                    output = network.run(population=population, seed=seed,
                                         s=s, nsteps=nsteps,
                                         output_dir=subdir,
                                         iterator=iterator,
                                         extractor=extractor,
                                         profile=profile,
                                         nthreads=nthreads)

                    outputs.append((variable, output))

                print(f"Completed job {i+1} of {len(variables)}")
                print(variable)
                print(output[-1])
            # end of OutputDirs context manager
        # end of loop over variable sets
    else:
        from ._worker import run_worker

        # create all of the parameters and options to run
        arguments = []

        for i, variable in enumerate(variables):
            seed = seeds[i]
            outdir = outdirs[i]

            arguments.append({
                "params": network.params.set_variables(variable),
                "options": {"seed": seed,
                            "output_dir": outdir,
                            "auto_bzip": output_dir.auto_bzip(),
                            "population": population,
                            "s": s,
                            "nsteps": nsteps,
                            "iterator": iterator,
                            "extractor": extractor,
                            "profile": profile,
                            "nthreads": nthreads}
            })

        if parallel_scheme == "multiprocessing":
            # run jobs using a multiprocessing pool
            print("\nRunning jobs in parallel using a multiprocessing pool...")
            from multiprocessing import Pool
            with Pool(processes=nprocs) as pool:
                results = pool.map(run_worker, arguments)

                for i, result in enumerate(results):
                    print(f"\nCompleted job {i+1} of {len(variables)}")
                    print(variables[i])
                    print(result[-1])
                    outputs.append((variables[i], result))

        elif parallel_scheme == "mpi4py":
            # run jobs using a mpi4py pool
            print("\nRunning jobs in parallel using a mpi4py pool...")
            from mpi4py import futures
            with futures.MPIPoolExecutor(max_workers=nprocs) as pool:
                results = pool.map(run_worker, arguments)

                for i, result in enumerate(results):
                    print(f"\nCompleted job {i+1} of {len(variables)}")
                    print(variables[i])
                    print(result[-1])
                    outputs.append((variables[i], result))

        elif parallel_scheme == "scoop":
            # run jobs using a scoop pool
            print("\nRunning jobs in parallel using a scoop pool...")
            from scoop import futures

            results = futures.map(run_worker, arguments)

            for i, result in enumerate(results):
                print(f"\nCompleted job {i+1} of {len(variables)}")
                print(variables[i])
                print(result[-1])
                outputs.append((variables[i], result))

        else:
            raise ValueError(f"Unrecognised parallelisation scheme "
                             f"{parallel_scheme}.")

    return outputs
